import logging
import os

import ibmsecurity
from ibmsecurity.utilities import tools

logger = logging.getLogger(__name__)

uri = "/wga/apiac/resource/instance"
requires_modules = ["wga"]
requires_version = "9.0.7"

try:
    basestring
except NameError:
    basestring = (str, bytes)


def get_all(isamAppliance, instance_name, check_mode=False, force=False):
    """
    Retrieve a list of all API Access Control Resource Servers
    """
    return isamAppliance.invoke_get("Retrieve a list of all API Access Control Resource Servers",
                                    "{0}/{1}/server".format(uri, instance_name),
                                    requires_modules=requires_modules, requires_version=requires_version)


def get(isamAppliance, instance_name, resource_server_name, check_mode=False, force=False):
    """
    Retrieve a single API Access Control Resource Server
    """
    return isamAppliance.invoke_get("Retrieve a single API Access Control Resource Server",
                                    "{0}/{1}/server{2}".format(uri, instance_name, resource_server_name),
                                    requires_modules=requires_modules, requires_version=requires_version)


def add(isamAppliance, instance_name, server_hostname, junction_point, junction_type, policy, authentication,
        static_response_headers=None, jwt=None, junction_hard_limit=None, junction_soft_limit=None,
        basic_auth_mode=None, tfim_sso=None, remote_http_header=None, stateful_junction=None,
        http2_junction=None, http2_proxy=None, sni_name=None, preserve_cookie=None, cookie_include_path=None,
        transparent_path_junction=None, mutual_auth=None, insert_ltpa_cookies=None,
        insert_session_cookies=None, request_encoding=None, enable_basic_auth=None, key_label=None,
        gso_resource_group=None, junction_cookie_javascript_block=None, client_ip_http=None,
        version_two_cookies=None, ltpa_keyfile=None, authz_rules=None, fsso_config_file=None, username=None,
        password=None, server_uuid=None, server_port=None, virtual_hostname=None, server_dn=None, local_ip=None,
        query_contents=None, case_sensitive_url=None, windows_style_url=None, ltpa_keyfile_password=None,
        https_port=None, http_port=None, proxy_hostname=None, proxy_port=None, sms_environment=None,
        vhost_label=None, junction_force=None, delegation_support=None, scripting_support=None,
        check_mode=False, force=False):
    """
    Creating a new API Access Control Resource Server
    """
    server_exist, warnings = _check_server_exist(isamAppliance, instance_name, junction_point)

    if force is True or server_exist is False:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            json_data = {
                'instance_name': instance_name,
                'server_hostname': server_hostname,
                'junction_point': junction_point,
                'junction_type': junction_type,
                'policy': policy,
                'authentication': authentication
            }
            if static_response_headers is not None:
                json_data['static_response_headers'] = static_response_headers

            if jwt is not None:
                json_data['jwt'] = jwt

            if junction_hard_limit is not None:
                json_data['junction_hard_limit'] = junction_hard_limit

            if junction_soft_limit is not None:
                json_data['junction_soft_limit'] = junction_soft_limit

            if basic_auth_mode is not None:
                json_data['basic_auth_mode'] = basic_auth_mode

            if tfim_sso is not None:
                json_data['tfim_sso'] = tfim_sso

            if remote_http_header is not None:
                json_data['remote_http_header'] = remote_http_header

            if stateful_junction is not None:
                json_data['stateful_junction'] = stateful_junction

            if http2_junction is not None:
                json_data['http2_junction']

            if http2_proxy is not None:
                json_data['http2_proxy'] = http2_proxy

            if sni_name is not None:
                json_data['sni_name'] = sni_name

            if preserve_cookie is not None:
                json_data['preserve_cookie'] = preserve_cookie

            if cookie_include_path is not None:
                json_data['cookie_include_path'] = cookie_include_path

            if transparent_path_junction is not None:
                json_data['transparent_path_junction'] = transparent_path_junction

            if mutual_auth is not None:
                json_data['mutual_auth'] = mutual_auth

            if insert_ltpa_cookies is not None:
                json_data['insert_ltpa_cookies'] = insert_ltpa_cookies

            if insert_session_cookies is not None:
                json_data['insert_session_cookies'] = insert_session_cookies

            if request_encoding is not None:
                json_data['request_encoding'] = request_encoding

            if enable_basic_auth is not None:
                json_data['enable_basic_auth'] = enable_basic_auth

            if key_label is not None:
                json_data['key_label'] = key_label

            if gso_resource_group is not None:
                json_data['gso_resource_group'] = gso_resource_group

            if junction_cookie_javascript_block is not None:
                json_data['junction_cookie_javascript_block'] = junction_cookie_javascript_block

            if client_ip_http is not None:
                json_data['client_ip_http'] = client_ip_http

            if version_two_cookies is not None:
                json_data['version_two_cookies'] = version_two_cookies

            if ltpa_keyfile is not None:
                json_data[ltpa_keyfile] = ltpa_keyfile

            if authz_rules is not None:
                json_data['authz_rules'] = authz_rules

            if fsso_config_file is not None:
                json_data['fsso_config_file'] = fsso_config_file

            if username is not None:
                json_data['username'] = username

            if password is not None:
                json_data['password'] = password

            if server_uuid is not None:
                json_data['server_uuid'] = server_uuid

            if server_port is not None:
                json_data['server_port'] = server_port

            if virtual_hostname is not None:
                json_data['virtual_hostname'] = virtual_hostname

            if server_dn is not None:
                json_data['server_dn'] = server_dn

            if local_ip is not None:
                json_data['local_ip'] = local_ip

            if query_contents is not None:
                json_data['query_contents'] = query_contents

            if case_sensitive_url is not None:
                json_data['case_sensitive_url'] = case_sensitive_url

            if windows_style_url is not None:
                json_data['windows_style_url'] = windows_style_url

            if ltpa_keyfile_password is not None:
                json_data['ltpa_keyfile_password'] = ltpa_keyfile_password

            if https_port is not None:
                json_data['https_port'] = https_port

            if http_port is not None:
                json_data['http_port'] = http_port

            if proxy_hostname is not None:
                json_data['proxy_hostname'] = proxy_hostname

            if proxy_port is not None:
                json_data['proxy_port'] = proxy_port

            if sms_environment is not None:
                json_data['sms_environment'] = sms_environment

            if vhost_label is not None:
                json_data['vhost_label'] = vhost_label

            if junction_force is not None:
                json_data['force'] = junction_force

            if delegation_support is not None:
                json_data['delegation_support'] = delegation_support

            if scripting_support is not None:
                json_data['scripting_support'] = scripting_support

            return isamAppliance.invoke_post(
                "Creating a new API Access Control Resource Server",
                "{0}/{1}/server".format(uri, instance_name),
                json_data, requires_modules=requires_modules, requires_version=requires_version)

    return isamAppliance.create_return_object(warnings=warnings)


def update(isamAppliance, instance_name, junction_point, server_hostname, junction_type,
           policy, authentication, server_type="standard", static_response_headers=None, jwt=None,
           junction_hard_limit=None, junction_soft_limit=None, basic_auth_mode=None, tfim_sso=None,
           remote_http_header=None, stateful_junction=None, http2_junction=None, http2_proxy=None,
           sni_name=None, preserve_cookie=None, cookie_include_path=None, transparent_path_junction=None,
           mutual_auth=None, insert_ltpa_cookies=None, insert_session_cookies=None, request_encoding=None,
           enable_basic_auth=None, key_label=None, gso_resource_group=None,
           junction_cookie_javascript_block=None, client_ip_http=None, version_two_cookies=None,
           ltpa_keyfile=None, authz_rules=None, fsso_config_file=None, username=None, password=None,
           server_uuid=None, server_port=None, virtual_hostname=None, server_dn=None, local_ip=None,
           query_contents=None, case_sensitive_url=None, windows_style_url=None, ltpa_keyfile_password=None,
           https_port=None, http_port=None, proxy_hostname=None, proxy_port=None, sms_environment=None,
           vhost_label=None, junction_force=True, delegation_support=None, scripting_support=None,
           check_mode=False, force=False):
    """
    Updating an existing API Access Control Resource Server
    """

    server_exist, warnings = _check_server_exist(isamAppliance, instance_name, junction_point)
    if server_exist is True:
        update_required, warnings = _check_server_content(isamAppliance=isamAppliance,
                                                          instance_name=instance_name,
                                                          junction_point=junction_point,
                                                          server_hostname=server_hostname,
                                                          junction_type=junction_type, policy=policy,
                                                          authentication=authentication,
                                                          static_response_headers=static_response_headers,
                                                          jwt=jwt,
                                                          junction_hard_limit=junction_hard_limit,
                                                          junction_soft_limit=junction_soft_limit,
                                                          basic_auth_mode=basic_auth_mode,
                                                          tfim_sso=tfim_sso,
                                                          remote_http_header=remote_http_header,
                                                          stateful_junction=stateful_junction,
                                                          http2_junction=http2_junction,
                                                          http2_proxy=http2_proxy, sni_name=sni_name,
                                                          preserve_cookie=preserve_cookie,
                                                          cookie_include_path=cookie_include_path,
                                                          transparent_path_junction=transparent_path_junction,
                                                          mutual_auth=mutual_auth,
                                                          insert_ltpa_cookies=insert_ltpa_cookies,
                                                          insert_session_cookies=insert_session_cookies,
                                                          request_encoding=request_encoding,
                                                          enable_basic_auth=enable_basic_auth,
                                                          key_label=key_label,
                                                          gso_resource_group=gso_resource_group,
                                                          junction_cookie_javascript_block=junction_cookie_javascript_block,
                                                          client_ip_http=client_ip_http,
                                                          version_two_cookies=version_two_cookies,
                                                          ltpa_keyfile=ltpa_keyfile, authz_rules=authz_rules,
                                                          fsso_config_file=fsso_config_file,
                                                          username=username,
                                                          password=password,
                                                          server_uuid=server_uuid, server_port=server_port,
                                                          virtual_hostname=virtual_hostname,
                                                          server_dn=server_dn,
                                                          local_ip=local_ip, query_contents=query_contents,
                                                          case_sensitive_url=case_sensitive_url,
                                                          windows_style_url=windows_style_url,
                                                          ltpa_keyfile_password=ltpa_keyfile_password,
                                                          https_port=https_port,
                                                          http_port=http_port, proxy_hostname=proxy_hostname,
                                                          proxy_port=proxy_port,
                                                          sms_environment=sms_environment,
                                                          vhost_label=vhost_label,
                                                          delegation_support=delegation_support,
                                                          scripting_support=scripting_support)

    else:
        return isamAppliance.create_return_object(warnings=warnings)

    if force is True or update_required is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            json_data = {
                'instance_name': instance_name,
                'server_hostname': server_hostname,
                'junction_point': junction_point,
                'junction_type': junction_type,
                'policy': policy,
                'authentication': authentication
            }
            if static_response_headers is not None:
                json_data['static_response_headers'] = static_response_headers

            if jwt is not None:
                json_data['jwt'] = jwt

            if junction_hard_limit is not None:
                json_data['junction_hard_limit'] = junction_hard_limit

            if junction_soft_limit is not None:
                json_data['junction_soft_limit'] = junction_soft_limit

            if basic_auth_mode is not None:
                json_data['basic_auth_mode'] = basic_auth_mode

            if tfim_sso is not None:
                json_data['tfim_sso'] = tfim_sso

            if remote_http_header is not None:
                json_data['remote_http_header'] = remote_http_header

            if stateful_junction is not None:
                json_data['stateful_junction'] = stateful_junction

            if http2_junction is not None:
                json_data['http2_junction']

            if http2_proxy is not None:
                json_data['http2_proxy'] = http2_proxy

            if sni_name is not None:
                json_data['sni_name'] = sni_name

            if preserve_cookie is not None:
                json_data['preserve_cookie'] = preserve_cookie

            if cookie_include_path is not None:
                json_data['cookie_include_path'] = cookie_include_path

            if transparent_path_junction is not None:
                json_data['transparent_path_junction'] = transparent_path_junction

            if mutual_auth is not None:
                json_data['mutual_auth'] = mutual_auth

            if insert_ltpa_cookies is not None:
                json_data['insert_ltpa_cookies'] = insert_ltpa_cookies

            if insert_session_cookies is not None:
                json_data['insert_session_cookies'] = insert_session_cookies

            if request_encoding is not None:
                json_data['request_encoding'] = request_encoding

            if enable_basic_auth is not None:
                json_data['enable_basic_auth'] = enable_basic_auth

            if key_label is not None:
                json_data['key_label'] = key_label

            if gso_resource_group is not None:
                json_data['gso_resource_group'] = gso_resource_group

            if junction_cookie_javascript_block is not None:
                json_data['junction_cookie_javascript_block'] = junction_cookie_javascript_block

            if client_ip_http is not None:
                json_data['client_ip_http'] = client_ip_http

            if version_two_cookies is not None:
                json_data['version_two_cookies'] = version_two_cookies

            if ltpa_keyfile is not None:
                json_data[ltpa_keyfile] = ltpa_keyfile

            if authz_rules is not None:
                json_data['authz_rules'] = authz_rules

            if fsso_config_file is not None:
                json_data['fsso_config_file'] = fsso_config_file

            if username is not None:
                json_data['username'] = username

            if password is not None:
                json_data['password'] = password

            if server_uuid is not None:
                json_data['server_uuid'] = server_uuid

            if server_port is not None:
                json_data['server_port'] = server_port

            if virtual_hostname is not None:
                json_data['virtual_hostname'] = virtual_hostname

            if server_dn is not None:
                json_data['server_dn'] = server_dn

            if local_ip is not None:
                json_data['local_ip'] = local_ip

            if query_contents is not None:
                json_data['query_contents'] = query_contents

            if case_sensitive_url is not None:
                json_data['case_sensitive_url'] = case_sensitive_url

            if windows_style_url is not None:
                json_data['windows_style_url'] = windows_style_url

            if ltpa_keyfile_password is not None:
                json_data['ltpa_keyfile_password'] = ltpa_keyfile_password

            if https_port is not None:
                json_data['https_port'] = https_port

            if http_port is not None:
                json_data['http_port'] = http_port

            if proxy_hostname is not None:
                json_data['proxy_hostname'] = proxy_hostname

            if proxy_port is not None:
                json_data['proxy_port'] = proxy_port

            if sms_environment is not None:
                json_data['sms_environment'] = sms_environment

            if vhost_label is not None:
                json_data['vhost_label'] = vhost_label

            if junction_force is not None:
                json_data['force'] = junction_force

            if delegation_support is not None:
                json_data['delegation_support'] = delegation_support

            if scripting_support is not None:
                json_data['scripting_support'] = scripting_support

            return isamAppliance.invoke_put(
                "Updating an existing API Access Control Resource Server",
                "{0}/{1}/server{2}?{3}".format(uri, instance_name, junction_point, server_type),
                json_data, requires_modules=requires_modules, requires_version=requires_version)

    return isamAppliance.create_return_object(warnings=warnings)


def set(isamAppliance, instance_name, junction_point, server_hostname, junction_type,
        policy, authentication, server_type="standard", static_response_headers=None, jwt=None,
        junction_hard_limit=None, junction_soft_limit=None, basic_auth_mode=None, tfim_sso=None,
        remote_http_header=None, stateful_junction=None, http2_junction=None, http2_proxy=None,
        sni_name=None, preserve_cookie=None, cookie_include_path=None, transparent_path_junction=None,
        mutual_auth=None, insert_ltpa_cookies=None, insert_session_cookies=None, request_encoding=None,
        enable_basic_auth=None, key_label=None, gso_resource_group=None,
        junction_cookie_javascript_block=None, client_ip_http=None, version_two_cookies=None,
        ltpa_keyfile=None, authz_rules=None, fsso_config_file=None, username=None, password=None,
        server_uuid=None, server_port=None, virtual_hostname=None, server_dn=None, local_ip=None,
        query_contents=None, case_sensitive_url=None, windows_style_url=None, ltpa_keyfile_password=None,
        https_port=None, http_port=None, proxy_hostname=None, proxy_port=None, sms_environment=None,
        vhost_label=None, junction_force=True, delegation_support=None, scripting_support=None,
        check_mode=False, force=False):
    server_exist, warnings = _check_server_exist(isamAppliance, instance_name, junction_point)

    if server_exist is True:
        return update(isamAppliance=isamAppliance, instance_name=instance_name, junction_point=junction_point,
                      server_hostname=server_hostname, junction_type=junction_type, policy=policy,
                      authentication=authentication, server_type=server_type,
                      static_response_headers=static_response_headers, jwt=jwt,
                      junction_hard_limit=junction_hard_limit,
                      junction_soft_limit=junction_soft_limit, basic_auth_mode=basic_auth_mode,
                      tfim_sso=tfim_sso, remote_http_header=remote_http_header,
                      stateful_junction=stateful_junction, http2_junction=http2_junction,
                      http2_proxy=http2_proxy, sni_name=sni_name, preserve_cookie=preserve_cookie,
                      cookie_include_path=cookie_include_path,
                      transparent_path_junction=transparent_path_junction, mutual_auth=mutual_auth,
                      insert_ltpa_cookies=insert_ltpa_cookies, insert_session_cookies=insert_session_cookies,
                      request_encoding=request_encoding, enable_basic_auth=enable_basic_auth,
                      key_label=key_label, gso_resource_group=gso_resource_group,
                      junction_cookie_javascript_block=junction_cookie_javascript_block,
                      client_ip_http=client_ip_http, version_two_cookies=version_two_cookies,
                      ltpa_keyfile=ltpa_keyfile, authz_rules=authz_rules, fsso_config_file=fsso_config_file,
                      username=username, password=password, server_uuid=server_uuid, server_port=server_port,
                      virtual_hostname=virtual_hostname, server_dn=server_dn, local_ip=local_ip,
                      query_contents=query_contents, case_sensitive_url=case_sensitive_url,
                      windows_style_url=windows_style_url, ltpa_keyfile_password=ltpa_keyfile_password,
                      https_port=https_port, http_port=http_port, proxy_hostname=proxy_hostname,
                      proxy_port=proxy_port, sms_environment=sms_environment, vhost_label=vhost_label,
                      junction_force=junction_force, delegation_support=delegation_support,
                      scripting_support=scripting_support, check_mode=check_mode, force=force)
    else:
        return add(isamAppliance=isamAppliance, instance_name=instance_name, server_hostname=server_hostname,
                   junction_point=junction_point, junction_type=junction_type, policy=policy,
                   authentication=authentication, static_response_headers=static_response_headers, jwt=jwt,
                   junction_hard_limit=junction_hard_limit, junction_soft_limit=junction_soft_limit,
                   basic_auth_mode=basic_auth_mode, tfim_sso=tfim_sso, remote_http_header=remote_http_header,
                   stateful_junction=stateful_junction, http2_junction=http2_junction, http2_proxy=http2_proxy,
                   sni_name=sni_name, preserve_cookie=preserve_cookie, cookie_include_path=cookie_include_path,
                   transparent_path_junction=transparent_path_junction, mutual_auth=mutual_auth,
                   insert_ltpa_cookies=insert_ltpa_cookies, insert_session_cookies=insert_session_cookies,
                   request_encoding=request_encoding, enable_basic_auth=enable_basic_auth, key_label=key_label,
                   gso_resource_group=gso_resource_group,
                   junction_cookie_javascript_block=junction_cookie_javascript_block,
                   client_ip_http=client_ip_http, version_two_cookies=version_two_cookies,
                   ltpa_keyfile=ltpa_keyfile, authz_rules=authz_rules, fsso_config_file=fsso_config_file,
                   username=username, password=password, server_uuid=server_uuid, server_port=server_port,
                   virtual_hostname=virtual_hostname, server_dn=server_dn, local_ip=local_ip,
                   query_contents=query_contents, case_sensitive_url=case_sensitive_url,
                   windows_style_url=windows_style_url, ltpa_keyfile_password=ltpa_keyfile_password,
                   https_port=https_port, http_port=http_port, proxy_hostname=proxy_hostname,
                   proxy_port=proxy_port, sms_environment=sms_environment, vhost_label=vhost_label,
                   junction_force=junction_force, delegation_support=delegation_support,
                   scripting_support=scripting_support, check_mode=check_mode, force=force)


def delete(isamAppliance, instance_name, resource_server_name, server_type='standard',
           check_mode=False, force=False):
    """
    Delete an existing API Access Control Resource Server
    """

    server_exist, warnings = _check_server_exist(isamAppliance, instance_name, resource_server_name)

    if force is True or server_exist is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            url = "{0}/{1}/server{2}?server_type={3}".format(uri, instance_name, resource_server_name, server_type)
            return isamAppliance.invoke_delete(
                "Delete an existing API Access Control Resource Server",
                url,
                requires_modules=requires_modules,
                requires_version=requires_version)

    return isamAppliance.create_return_object(warnings=warnings)


def delete_selection(isamAppliance, instance_name, resource_servers, command="DELETE",
                     check_mode=False, force=False):
    """
    Delete a selection of API Access Control Resource Servers
    """

    ret_obj = get_all(isamAppliance, instance_name)
    warnings = ret_obj['warnings']
    found_any = False
    new_list_servers = []

    for server in resource_servers:
        found = False
        for obj in ret_obj['data']:
            if obj['name'] == server:
                found = True
                found_any = True
        if found is False:
            warnings.append("Did not find resource server {0} to delete".format(server))
        else:
            new_list_servers.append(server)

    if force is True or found_any is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_put(
                "Delete a selection of API Access Control Resource Servers",
                "{0}/{1}/server".format(uri, instance_name),
                {
                    'command': command,
                    'resource_servers': new_list_servers
                },
                requires_modules=requires_modules,
                requires_version=requires_version, warnings=warnings)

    return isamAppliance.create_return_object(warnings=warnings)


def delete_all(isamAppliance, instance_name, check_mode=False, force=False):
    """
    Delete all existing API Access Control Resource Servers
    """

    all_exist, warnings = _check_all_servers(isamAppliance, instance_name)

    if force is True or all_exist is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            url = "{0}/{1}/server".format(uri, instance_name)
            return isamAppliance.invoke_delete(
                "Delete all existing API Access Control Resource Servers",
                url,
                requires_modules=requires_modules,
                requires_version=requires_version)

    return isamAppliance.create_return_object(warnings=warnings)


def export_all(isamAppliance, instance_name, file_path, check_mode=False, force=False):
    """
    Exporting all existing API Access Control Resource Servers
    """

    if os.path.exists(file_path) is True:
        warn_str = "File {0} already exists".format(file_path)
        warnings = [warn_str]
        return isamAppliance.create_return_object(warnings=warnings)

    instance_exist, warnings = _check_instance_exist(isamAppliance, instance_name)

    if force is True or instance_exist is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            url = "{0}/{1}/server?export=true".format(uri, instance_name)
            return isamAppliance.invoke_get_file(
                "Exporting all existing API Access Control Resource Servers",
                url,
                file_path,
                requires_modules=requires_modules,
                requires_version=requires_version)

    return isamAppliance.create_return_object(warnings=warnings)


def export_file(isamAppliance, instance_name, resource_server_name, file_path, check_mode=False, force=False):
    """
    Exporting an existing API Access Control Resource Server
    """
    if os.path.exists(file_path) is True:
        warn_str = "File {0} already exists".format(file_path)
        warnings = [warn_str]
        return isamAppliance.create_return_object(warnings=warnings)

    instance_exist, warnings = _check_instance_exist(isamAppliance, instance_name)
    if force is False:
        if instance_exist is False:
            warnings.append("{0} does not exist".format(instance_name))
            return isamAppliance.create_return_object(warnings=warnings)
        else:
            server_exist, warnings = _check_server_exist(isamAppliance, instance_name, resource_server_name)
            if server_exist is False:
                warnings.append("The specified resource server name does not exist")
                return isamAppliance.create_return_object(warnings=warnings)

    if force is True or server_exist is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            url = "{0}/{1}/server{2}?export=true".format(uri, instance_name, resource_server_name)
            return isamAppliance.invoke_get_file(
                "Exporting an existing API Access Control Resource Server",
                url,
                file_path,
                requires_modules=requires_modules,
                requires_version=requires_version)

    return isamAppliance.create_return_object(warnings=warnings)


def import_file(isamAppliance, instance_name, filename, check_mode=False, force=False):
    """
    Importing an API Access Control Resource Server(s)
    """
    if os.path.exists(filename) is False:
        warn_str = "File {0} does not exists".format(filename)
        warnings = [warn_str]
        return isamAppliance.create_return_object(warnings=warnings)

    instance_exist, warnings = _check_instance_exist(isamAppliance, instance_name)
    if force is False:
        if instance_exist is False:
            warnings.append("{0} does not exist".format(instance_name))
            return isamAppliance.create_return_object(warnings=warnings)

    if force is True or instance_exist is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            url = "{0}/{1}/server".format(uri, instance_name)

            return isamAppliance.invoke_post_files(
                "Importing an API Access Control Resource Server(s)", url,
                [
                    {
                        'file_formfield': 'config_file',
                        'filename': filename,
                        'mimetype': 'application/octet-stream'
                    }
                ],
                {}, requires_modules=requires_modules, requires_version=requires_version)

    return isamAppliance.create_return_object(warnings=warnings)


def _check_server_exist(isamAppliance, instance_name, junction_point):
    ret_obj = get_all(isamAppliance, instance_name)
    for obj in ret_obj['data']:
        if obj['name'] == junction_point:
            return True, ret_obj['warnings']

    return False, ret_obj['warnings']


def _check_all_servers(isamAppliance, instance_name):
    ret_obj = get_all(isamAppliance, instance_name)
    warnings = ret_obj['warnings']

    if ret_obj['data'] != []:
        return True, warnings
    else:
        return False, warnings


def _check_list_servers(isamAppliance, instance_name, resource_servers):
    ret_obj = get_all(isamAppliance, instance_name)
    warnings = ret_obj['warnings']
    non_exist = False

    for server in resource_servers:
        found = False
        for obj in ret_obj['data']:
            if obj['name'] == server:
                found = True
        if found is False:
            non_exist = True
            warnings.append("Did not find resource server {0}".format(server))

    if non_exist is False:
        return True, ret_obj['warnings']
    else:
        return False, warnings


def _check_server_content(isamAppliance, instance_name, junction_point, server_hostname, junction_type, policy,
                          authentication, static_response_headers, jwt, junction_hard_limit, junction_soft_limit,
                          basic_auth_mode, tfim_sso, remote_http_header, stateful_junction, http2_junction, http2_proxy,
                          sni_name, preserve_cookie, cookie_include_path, transparent_path_junction, mutual_auth,
                          insert_ltpa_cookies, insert_session_cookies, request_encoding, enable_basic_auth, key_label,
                          gso_resource_group, junction_cookie_javascript_block, client_ip_http, version_two_cookies,
                          ltpa_keyfile, authz_rules, fsso_config_file, username, password, server_uuid, server_port,
                          virtual_hostname, server_dn, local_ip, query_contents, case_sensitive_url,
                          windows_style_url, ltpa_keyfile_password, https_port, http_port, proxy_hostname, proxy_port,
                          sms_environment, vhost_label, delegation_support, scripting_support):
    ret_obj = get(isamAppliance, instance_name, junction_point)
    current_data = ret_obj['data']
    add_required = False
    json_data = {
        'server_hostname': server_hostname,
        'junction_point': junction_point,
        'junction_type': junction_type.lower(),
        'policy': policy,
        'authentication': authentication
    }

    server_str = current_data['servers']
    server_list = server_str.split("#")

    for server in server_list:
        if server[-1] == ";":
            str_length = len(server)
            server_str = server[:str_length - 1]
        else:
            server_str = server
        server_values = server_str.split(";")
        srv = {}
        for values in server_values:
            jnames = values.split("!")
            value_pair = {
                jnames[0]: jnames[1]
            }
            srv.update(value_pair)
        server_found = False
        if srv['server_hostname'] == server_hostname and str(srv['server_port']) == str(server_port):
            logger.debug("Matched a server - {0}.".format(srv))
            server_found = True
            server_json = {
                'server_hostname': server_hostname,
                'server_port': str(server_port)
            }
            if case_sensitive_url is None:
                server_json['case_sensitive_url'] = 'no'
            else:
                server_json['case_sensitive_url'] = case_sensitive_url

            if http_port is None:
                server_json['http_port'] = str(server_port)
            else:
                server_json['http_port'] = str(http_port)

            if https_port is not None:
                server_json['https_port'] = str(https_port)

            if local_ip is None:
                server_json['local_ip'] = ''
            else:
                server_json['local_ip'] = local_ip

            if query_contents is None:
                server_json['query_content_url'] = '/cgi-bin/query_contents'
            else:
                server_json['query_content_url'] = query_contents

            if server_dn is None:
                server_json['server_dn'] = ''
            else:
                server_json['server_dn'] = server_dn

            if server_uuid is not None:
                server_json['server_uuid'] = server_uuid
            else:
                # Server UUID gets generated if not specified
                if 'server_uuid' in srv:
                    del srv['server_uuid']

            if virtual_hostname is None:
                if 'virtual_junction_hostname' in srv:
                    del srv['virtual_junction_hostname']
            else:
                server_json['virtual_junction_hostname'] = virtual_hostname

            if windows_style_url is None:
                server_json['windows_style_url'] = 'no'
            else:
                server_json['windows_style_url'] = windows_style_url

            # Delete dynamic data shown when we get junctions details
            if 'current_requests' in srv:
                del srv['current_requests']

            if 'total_requests' in srv:
                del srv['total_requests']

            if 'operation_state' in srv:
                del srv['operation_state']

            if 'server_state' in srv:
                del srv['server_state']

            # Not sure what this attribute is supposed to contain?
            if 'query_contents' in srv:
                del srv['query_contents']

            if tools.json_sort(server_json) != tools.json_sort(srv):
                logger.debug("Servers are found to be different. See following JSON for difference.")
                logger.debug("New Server JSON: {0}".format(tools.json_sort(server_json)))
                logger.debug("Old Server JSON: {0}".format(tools.json_sort(srv)))
                add_required = True
            break

    if server_found is False:
        add_required = True
    elif add_required is False:
        if static_response_headers is not None:
            json_data['static_response_headers'] = static_response_headers
        else:
            json_data['static_response_headers'] = {}

        if jwt is not None:
            json_data['jwt'] = jwt
        else:
            json_data['jwt'] = {}

        if junction_hard_limit is not None:
            json_data['junction_hard_limit'] = junction_hard_limit
        else:
            json_data['junction_hard_limit'] = '0 - using global value'

        if junction_soft_limit is not None:
            json_data['junction_soft_limit'] = junction_soft_limit
        else:
            json_data['junction_soft_limit'] = '0 - using global value',

        if basic_auth_mode is not None:
            json_data['basic_auth_mode'] = basic_auth_mode
        else:
            json_data['basic_auth_mode'] = 'filter'

        if tfim_sso is not None:
            json_data['tfim_sso'] = tfim_sso
        else:
            json_data['tfim_sso'] = 'no'

        if remote_http_header is None:
            json_data['remote_http_header'] = 'do not insert'
        elif isinstance(remote_http_header, basestring) and remote_http_header.lower() == 'all':
            json_data['remote_http_header'] = ['iv_creds', 'iv_groups', 'iv_user']
        else:
            json_data['remote_http_header'] = remote_http_header

        if 'remote_http_header' in ret_obj['data']:
            http_headers = ret_obj['data']['remote_http_header'].split(' ')
            if 'insert' in http_headers:
                http_headers.remove('insert')
            if '-' in http_headers:
                http_headers.remove('-')
            current_data['remote_http_header'] = http_headers

        if stateful_junction is not None:
            json_data['stateful_junction'] = stateful_junction
        else:
            json_data['stateful_junction'] = 'no'

        if http2_junction is not None:
            json_data['http2_junction'] = http2_junction

        if http2_proxy is not None:
            json_data['http2_proxy'] = http2_proxy

        if sni_name is not None:
            json_data['sni_name'] = sni_name

        if preserve_cookie is not None:
            json_data['preserve_cookie'] = preserve_cookie
        else:
            json_data['preserve_cookie'] = 'no'

        if cookie_include_path is not None:
            json_data['cookie_include_path'] = cookie_include_path
        else:
            json_data['cookie_include_path'] = 'no'

        if transparent_path_junction is not None:
            json_data['transparent_path_junction'] = transparent_path_junction
        else:
            json_data['transparent_path_junction'] = 'no'

        if mutual_auth is not None:
            json_data['mutual_auth'] = mutual_auth
        else:
            json_data['mutual_auth'] = mutual_auth

        if insert_ltpa_cookies is not None:
            json_data['insert_ltpa_cookies'] = insert_ltpa_cookies

        if insert_session_cookies is not None:
            json_data['insert_session_cookies'] = insert_session_cookies
        else:
            json_data['insert_session_cookies'] = 'no'

        if request_encoding is not None:
            json_data['request_encoding'] = request_encoding
        else:
            json_data['request_encoding'] = 'utf8_uri'

        if 'request_encoding' in ret_obj['data']:
            if ret_obj['data']['request_encoding'] == "UTF-8, URI Encoded":
                current_data['request_encoding'] = 'utf8_uri'
            elif ret_obj['data']['request_encoding'] == "UTF-8, Binary":
                current_data['request_encoding'] = 'utf8_bin'
            elif ret_obj['data']['request_encoding'] == "Local Code Page, Binary":
                current_data['request_encoding'] = 'lcp_bin'
            elif ret_obj['data']['request_encoding'] == "Local Code Page, URI Encoded":
                current_data['request_encoding'] = 'lcp_uri'

        if enable_basic_auth is not None:
            json_data['enable_basic_auth'] = enable_basic_auth

        if key_label is not None:
            json_data['key_label'] = key_label

        if gso_resource_group is not None:
            json_data['gso_resource_group'] = gso_resource_group

        if junction_cookie_javascript_block is not None:
            json_data['junction_cookie_javascript_block'] = junction_cookie_javascript_block

        if client_ip_http is not None:
            json_data['client_ip_http'] = client_ip_http
            if 'client_ip_http' in ret_obj['data']:
                if ret_obj['data']['client_ip_http'] == 'do not insert':
                    current_data['client_ip_http'] = 'no'
                else:
                    current_data['client_ip_http'] = ret_obj['data']['client_ip_http']

        if client_ip_http is None or client_ip_http.lower() == 'no':
            json_data['client_ip_http'] = 'do not insert'
        elif client_ip_http.lower() == 'yes':
            json_data['client_ip_http'] = 'insert'
        else:
            json_data['client_ip_http'] = client_ip_http

        if version_two_cookies is not None:
            json_data['client_ip_http'] = client_ip_http

        if ltpa_keyfile is not None:
            json_data[ltpa_keyfile] = ltpa_keyfile

        if authz_rules is not None:
            json_data['authz_rules'] = authz_rules
        else:
            json_data['authz_rules'] = 'no'

        if fsso_config_file is not None:
            json_data['fsso_config_file'] = fsso_config_file
        else:
            json_data['fsso_config_file'] = 'disabled'

        if username is not None:
            json_data['username'] = username

        if password is not None:
            json_data['password'] = password

        if ltpa_keyfile_password is not None:
            json_data['ltpa_keyfile_password'] = ltpa_keyfile_password

        if proxy_hostname is not None:
            json_data['proxy_hostname'] = proxy_hostname

        if proxy_port is not None:
            json_data['proxy_port'] = proxy_port

        if sms_environment is not None:
            json_data['sms_environment'] = sms_environment

        if vhost_label is not None:
            json_data['vhost_label'] = vhost_label

        if delegation_support is not None:
            json_data['delegation_support'] = delegation_support
        else:
            json_data['delegation_support'] = 'no'

        if scripting_support is not None:
            json_data['scripting_support'] = scripting_support
        else:
            json_data['scripting_support'] = 'no'

        del current_data['boolean_rule_header']
        del current_data['forms_based_sso']
        del current_data['http_header_ident']
        del current_data['session_cookie_backend_portal']
        del current_data['servers']
        sorted_obj1 = tools.json_sort(json_data)
        logger.debug("Sorted sorted_obj1: {0}".format(sorted_obj1))
        sorted_obj2 = tools.json_sort(current_data)
        logger.debug("Sorted sorted_obj2: {0}".format(sorted_obj2))

        if sorted_obj1 != sorted_obj2:
            logger.info("Changes detected, update needed.")
            add_required = True

    return add_required, ret_obj['warnings']


def _check_instance_exist(isamAppliance, instance_name):
    ret_obj = ibmsecurity.isam.web.api_access_control.resources.instances.get_all(isamAppliance)

    for obj in ret_obj['data']:
        if obj['name'] == instance_name:
            return True, ret_obj['warnings']

    return False, ret_obj['warnings']


def compare(isamAppliance1, isamAppliance2):
    """
    Compare resources between two appliances
    """

    app1_instances = ibmsecurity.isam.web.api_access_control.resources.instances.get_all(isamAppliance1)
    app2_instances = ibmsecurity.isam.web.api_access_control.resources.instances.get_all(isamAppliance2)

    obj1 = []
    obj2 = []

    for inst1 in app1_instances['data']:
        servers = get_all(isamAppliance1, instance_name=inst1['name'])
        for srv in servers['data']:
            if "servers" in srv:
                srvlist = srv['servers'].split(";")
                new_str = None
                for value in srvlist:
                    if value.find("server_uuid") == -1 and \
                            value.find("current_requests") == -1 and \
                            value.find("total_requests") == -1:
                        if new_str is None:
                            new_str = value
                        else:
                            new_str = new_str + ";" + value
                srv['servers'] = new_str
            obj1.append(srv)

    for inst2 in app2_instances['data']:
        servers = get_all(isamAppliance2, instance_name=inst2['name'])
        for srv in servers['data']:
            if "servers" in srv:
                srvlist = srv['servers'].split(";")
                new_str = None
                for value in srvlist:
                    if value.find("server_uuid") == -1 and \
                            value.find("current_requests") == -1 and \
                            value.find("total_requests") == -1:
                        if new_str is None:
                            new_str = value
                        else:
                            new_str = new_str + ";" + value
                srv['servers'] = new_str
            obj2.append(srv)

    app1_instances['data'].extend(obj1)
    app2_instances['data'].extend(obj2)

    return tools.json_compare(app1_instances, app2_instances,
                              deleted_keys=['server_uuid', 'current_requests', 'total_requests'])


def compare_one_instance(isamAppliance1, isamAppliance2, instance1_name, instance2_name=None):
    """
    Compare resources between two appliances
    """
    if instance2_name is None or instance2_name == '':
        instance2_name = instance1_name

    obj1 = []
    obj2 = []
    servers1 = get_all(isamAppliance1, instance_name=instance1_name)
    for srv in servers1['data']:
        if "servers" in srv:
            srvlist = srv['servers'].split(";")
            new_str = None
            for value in srvlist:
                if value.find("server_uuid") == -1 and \
                        value.find("current_requests") == -1 and \
                        value.find("total_requests") == -1:
                    if new_str is None:
                        new_str = value
                    else:
                        new_str = new_str + ";" + value
            srv['servers'] = new_str
        obj1.append(srv)

    servers2 = get_all(isamAppliance2, instance_name=instance2_name)
    for srv in servers2['data']:
        if "servers" in srv:
            srvlist = srv['servers'].split(";")
            new_str = None
            for value in srvlist:
                if value.find("server_uuid") == -1 and \
                        value.find("current_requests") == -1 and \
                        value.find("total_requests") == -1:
                    if new_str is None:
                        new_str = value
                    else:
                        new_str = new_str + ";" + value
            srv['servers'] = new_str
        obj2.append(srv)

    servers1['data'] = obj1
    servers2['data'] = obj2

    return tools.json_compare(servers1, servers2, deleted_keys=['server_uuid', 'current_requests', 'total_requests'])



import logging
from ibmsecurity.utilities import tools
from io import open

logger = logging.getLogger(__name__)

# URI for this module
uri = "/iam/access/v8/policies"


def get_all(isamAppliance, filter=None, sortBy=None, check_mode=False, force=False):
    """
    Retrieve a list of policies
    """
    return isamAppliance.invoke_get("Retrieve a list of policies",
                                    "{0}/{1}".format(uri, tools.create_query_string(filter=filter, sortBy=sortBy)))


def get(isamAppliance, name, check_mode=False, force=False):
    """
    Retrieve a specific policy
    """
    ret_obj = search(isamAppliance, name=name, check_mode=check_mode, force=force)
    pol_id = ret_obj['data']

    if pol_id == {}:
        logger.info("Policy {0} had no match, skipping retrieval.".format(name))
        return isamAppliance.create_return_object()
    else:
        return _get(isamAppliance, pol_id)


def _get(isamAppliance, pol_id):
    return isamAppliance.invoke_get("Retrieve a specific policy",
                                    "{0}/{1}".format(uri, pol_id))


def export_xacml(isamAppliance, name, filename, check_mode=False, force=False):
    """
    Export XACML for a specific policy
    """
    import os.path
    if force is False:
        ret_obj = get(isamAppliance, name=name, check_mode=check_mode, force=force)

    if force is True or (ret_obj['data'] != {} and os.path.exists(filename) is False):
        if check_mode is False:  # No point downloading a file if in check_mode
            f = open(filename, 'w')
            f.write(ret_obj['data']['policy'])

    return isamAppliance.create_return_object()


def search(isamAppliance, name, force=False, check_mode=False):
    """
    Search policy id by name
    """
    ret_obj = get_all(isamAppliance)
    return_obj = isamAppliance.create_return_object()

    for obj in ret_obj['data']:
        if obj['name'] == name:
            logger.info("Found Policy {0} id: {1}".format(name, obj['id']))
            return_obj['data'] = obj['id']
            return_obj['rc'] = 0

    return return_obj


def set_file(isamAppliance, name, attributesrequired, policy_file, description="",
             dialect="urn:oasis:names:tc:xacml:2.0:policy:schema:os", predefined=False, new_name=None, check_mode=False,
             force=False):
    # Read policy from file and call set()
    with open(policy_file, 'r') as myfile:
        policy = myfile.read().replace('\n', '')
    return set(isamAppliance, name, attributesrequired, policy, description, dialect, predefined, new_name, check_mode,
               force)


def set(isamAppliance, name, attributesrequired, policy, description="",
        dialect="urn:oasis:names:tc:xacml:2.0:policy:schema:os", predefined=False, new_name=None, check_mode=False,
        force=False):
    """
    Creating or Modifying a Policy
    """
    if (search(isamAppliance, name=name))['data'] == {}:
        # Force the add - we already know policy does not exist
        logger.info("Policy {0} had no match, requesting to add new one.".format(name))
        return add(isamAppliance, name, attributesrequired, policy, description, dialect, predefined, check_mode, True)
    else:
        # Update request
        logger.info("Policy {0} exists, requesting to update.".format(name))
        return update(isamAppliance, name, attributesrequired, policy, description, dialect, predefined, new_name,
                      check_mode, force)


def add(isamAppliance, name, attributesrequired, policy, description="",
        dialect="urn:oasis:names:tc:xacml:2.0:policy:schema:os", predefined=False, check_mode=False, force=False):
    """
    Create a new Policy
    """
    if force is False:
        ret_obj = search(isamAppliance, name)

    if force is True or ret_obj['data'] == {}:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            try:
                import json
                json_data = json.loads(policy)[0]
                logger.info("Policy {0} contains full policy export".format(name))
            except json.decoder.JSONDecodeError:
                logger.info("Policy {0} only contains policy data".format(name))
                json_data = {
                    "name": name,
                    "attributesrequired": attributesrequired,
                    "description": description,
                    "predefined": predefined,
                    "policy": policy,
                    "dialect": dialect
                }

            return isamAppliance.invoke_post("Create a new Policy", uri, json_data)

    return isamAppliance.create_return_object()


def delete(isamAppliance, name, check_mode=False, force=False):
    """
    Delete a Policy
    """
    ret_obj = search(isamAppliance, name, check_mode=check_mode, force=force)
    mech_id = ret_obj['data']

    if mech_id == {}:
        logger.info("Policy {0} not found, skipping delete.".format(name))
    else:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_delete(
                "Delete a Policy",
                "{0}/{1}".format(uri, mech_id))

    return isamAppliance.create_return_object()


def update(isamAppliance, name, attributesrequired, policy, description=None,
           dialect="urn:oasis:names:tc:xacml:2.0:policy:schema:os", predefined=False, new_name=None, check_mode=False,
           force=False):
    """
    Update a specified policy
    """
    pol_id, update_required, json_data = _check(isamAppliance, name, attributesrequired, policy, description,
                                                dialect, predefined, new_name)
    if pol_id is None:
        from ibmsecurity.appliance.ibmappliance import IBMError
        raise IBMError("999", "Cannot update data for unknown policy: {0}".format(name))

    if force is True or update_required is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_put(
                "Update a specified policy",
                "{0}/{1}".format(uri, pol_id), json_data)

    return isamAppliance.create_return_object()


def _check(isamAppliance, name, attributesrequired, policy, description, dialect, predefined, new_name):
    """
    Check and return True if update needed
    """
    update_required = False
    try:
        import json
        json_data = json.loads(policy)[0]
        logger.info("Policy {0} contains full policy export".format(name))
        logger.info(json_data)
        if 'datecreated' in json_data:
            del json_data['datecreated']
        if 'id' in json_data:
            del json_data['id']
        if 'lastmodified' in json_data:
            del json_data['lastmodified']
        if 'userlastmodified' in json_data:
            del json_data['userlastmodified']
    except json.decoder.JSONDecodeError:
        logger.info("Policy {0} only contains policy data".format(name))
        json_data = {
            "attributesrequired": attributesrequired,
            "policy": policy,
            "dialect": dialect,
            "predefined": predefined
        }

    ret_obj = get(isamAppliance, name)
    if ret_obj['data'] == {}:
        logger.warning("Policy not found, returning no update required.")
        return None, update_required, json_data
    else:
        pol_id = ret_obj['data']['id']
        if new_name is not None:
            json_data['name'] = new_name
        else:
            json_data['name'] = name
        if description is not None:
            json_data['description'] = description
        else:
            del ret_obj['data']['description']
        del ret_obj['data']['id']
        del ret_obj['data']['userlastmodified']
        del ret_obj['data']['lastmodified']
        del ret_obj['data']['datecreated']
        import ibmsecurity.utilities.tools
        sorted_json_data = ibmsecurity.utilities.tools.json_sort(json_data)
        logger.debug("Sorted input: {0}".format(sorted_json_data))
        sorted_ret_obj = ibmsecurity.utilities.tools.json_sort(ret_obj['data'])
        logger.debug("Sorted existing data: {0}".format(sorted_ret_obj))
        if sorted_ret_obj != sorted_json_data:
            logger.info("Changes detected, update needed.")
            update_required = True

    return pol_id, update_required, json_data


def compare(isamAppliance1, isamAppliance2):
    """
    Compare Policies between two appliances
    """
    ret_obj1 = get_all(isamAppliance1)
    ret_obj2 = get_all(isamAppliance2)

    for obj in ret_obj1['data']:
        del obj['id']
        del obj['userlastmodified']
        del obj['lastmodified']
        del obj['datecreated']
        ret_obj = get(isamAppliance1, obj['name'])
        obj['policy'] = ret_obj['data']['policy']
    for obj in ret_obj2['data']:
        del obj['id']
        del obj['userlastmodified']
        del obj['lastmodified']
        del obj['datecreated']
        ret_obj = get(isamAppliance2, obj['name'])
        obj['policy'] = ret_obj['data']['policy']

    return tools.json_compare(ret_obj1, ret_obj2,
                              deleted_keys=['id', 'userlastmodified', 'lastmodified', 'datecreated'])



import json
import requests
from requests.packages.urllib3.exceptions import InsecureRequestWarning
import logging
from .ibmappliance import IBMAppliance
from .ibmappliance import IBMError
from ibmsecurity.utilities import tools
from io import open

try:
    basestring
except NameError:

    basestring = (str, bytes)


class ISVGAppliance(IBMAppliance):
    def __init__(self, hostname, user, lmi_port=443):
        self.logger = logging.getLogger(__name__)
        self.logger.debug('Creating an ISVGAppliance')
        if isinstance(lmi_port, basestring):
            self.lmi_port = int(lmi_port)
        else:
            self.lmi_port = lmi_port

        IBMAppliance.__init__(self, hostname, user)

    def _url(self, uri):
        # Build up the URL
        url = "https://" + self.hostname + ":" + str(self.lmi_port) + uri
        self.logger.debug("Issuing request to: " + url)

        return url

    def _log_desc(self, description):
        if description != "":
            self.logger.info('*** ' + description + ' ***')

    def _suppress_ssl_warning(self):
        # Disable https warning because of non-standard certs on appliance
        try:
            self.logger.debug("Suppressing SSL Warnings.")
            requests.packages.urllib3.disable_warnings(InsecureRequestWarning)
        except AttributeError:
            self.logger.warning("load requests.packages.urllib3.disable_warnings() failed")

    def _process_response(self, return_obj, http_response, ignore_error):

        return_obj['rc'] = http_response.status_code

        # Examine the response.
        if (http_response.status_code == 302):
            self.logger.warning("  Endpoint cannot be use at this time: ")
            self.logger.warning("     status code: {0}".format(http_response.status_code))
            if http_response.text != "":
                self.logger.error("     text: " + http_response.text)
            # Too early to use this endppoint.
            return_obj['changed'] = False
            return_obj['rc'] = 0
        elif (http_response.status_code != 200 and http_response.status_code != 204 and http_response.status_code != 201):
            self.logger.error("  Request failed: ")
            self.logger.error("     status code: {0}".format(http_response.status_code))
            if http_response.text != "":
                self.logger.error("     text: " + http_response.text)
            if not ignore_error:
                raise IBMError("HTTP Return code: {0}".format(http_response.status_code), http_response.text)
            return_obj['changed'] = False  # force changed to be False as there is an error
        else:
            return_obj['rc'] = 0

        # Handle if there was json on input but response was not in json format
        try:
            json_data = json.loads(http_response.text)
        except ValueError:
            return_obj['data'] = http_response.content
            return

        self.logger.debug("Status Code: {0}".format(http_response.status_code))
        if http_response.text != "":
            self.logger.debug("Text: " + http_response.content.decode("utf-8"))

        for key in http_response.headers:
            if key == 'g-type':
                if http_response.headers[key] == 'application/octet-stream; charset=UTF-8':
                    json_data = {}
                    return_obj.data = http_response.content
                    return

        if http_response.text == "":
            json_data = {}
        else:
            json_data = json.loads(http_response.text)

        return_obj['data'] = json_data

    def _process_connection_error(self, ignore_error, return_obj):
        if not ignore_error:
            self.logger.critical("Failed to connect to server.")
            raise IBMError("HTTP Return code: 502", "Failed to connect to server")
        else:
            self.logger.debug("Failed to connect to server.")
            return_obj['rc'] = 502

    def _process_warnings(self, uri, requires_modules, requires_version, warnings=[]):
        # flag to indicate if processing needs to return and not continue
        return_call = False
        self.logger.debug("Checking for minimum version: {0}.".format(requires_version))
        if requires_version is not None and 'version' in self.facts and self.facts['version'] is not None:
            if self.facts['version'] < requires_version:
                return_call = True
                warnings.append(
                    "API invoked requires minimum version: {0}, appliance is of lower version: {1}.".format(
                        requires_version, self.facts['version']))
        # Detecting modules from uri if none is provided
        if requires_modules is None and not requires_modules:
            if uri.startswith("/wga"):
                requires_modules = ['wga']
                self.logger.debug("Detected module: {0} from uri: {1}.".format(requires_modules, uri))
            elif uri.startswith("/mga"):
                requires_modules = ['mga']
                self.logger.debug("Detected module: {0} from uri: {1}.".format(requires_modules, uri))

        self.logger.debug("Checking for one of required modules: {0}.".format(requires_modules))
        if requires_modules is not None and requires_modules:
            if 'activations' in self.facts and self.facts['activations']:
                # Find intersection of the two lists
                iactive = [ia for ia in self.facts['activations'] if ia in requires_modules]
                if not iactive:
                    return_call = True
                    warnings.append(
                        "API invoked requires one of modules: {0}, appliance has these modules active: {1}.".format(
                            requires_modules, self.facts['activations']))
                else:
                    self.logger.info("Modules satisfying requirement: {0}".format(iactive))
            else:
                return_call = True
                warnings.append("API invoked requires module: {0}, appliance has no modules active.".format(
                    requires_modules))

        self.logger.debug("Warnings: {0}".format(warnings))
        return warnings, return_call

    def invoke_post_files(self, description, uri, fileinfo, data, ignore_error=False, requires_modules=None,
                          requires_version=None, warnings=[], json_response=True):
        """
        Send multipart/form-data upload file request to the appliance.
        """
        self._log_desc(description=description)

        warnings, return_call = self._process_warnings(uri=uri, requires_modules=requires_modules,
                                                       requires_version=requires_version,
                                                       warnings=warnings)
        return_obj = self.create_return_object(warnings=warnings)
        if return_call:
            return return_obj

        # Build up the URL and header information.
        if json_response:
            headers = {
                'Accept': 'application/json,text/html,application/xhtml+xml,application/xml'
            }
        else:
            headers = {
                'Accept': 'text/html,application/xhtml+xml,application/xml'
            }
        self.logger.debug("Headers are: {0}".format(headers))

        files = list()
        for file2post in fileinfo:
            files.append((file2post['file_formfield'],
                          (tools.path_leaf(file2post['filename']), open(file2post['filename'], 'rb'),
                           file2post['mimetype'])))

        self._suppress_ssl_warning()

        try:
            r = requests.post(url=self._url(uri=uri), data=data, auth=(self.user.username, self.user.password),
                              files=files, verify=False, headers=headers)
            return_obj['changed'] = True  # POST of file would be a change
            self._process_response(return_obj=return_obj, http_response=r, ignore_error=ignore_error)

        except requests.exceptions.ConnectionError:
            if not ignore_error:
                self.logger.critical("Failed to connect to server.")
                raise IBMError("HTTP Return code: 502", "Failed to connect to server")
            else:
                self.logger.debug("Failed to connect to server.")
                return_obj.rc = 502

        return return_obj

    def invoke_put_files(self, description, uri, fileinfo, data, ignore_error=False, requires_modules=None,
                         requires_version=None, warnings=[]):
        """
        Send multipart/form-data upload file request to the appliance.
        """
        self._log_desc(description=description)

        warnings, return_call = self._process_warnings(uri=uri, requires_modules=requires_modules,
                                                       requires_version=requires_version,
                                                       warnings=warnings)
        return_obj = self.create_return_object(warnings=warnings)
        if return_call:
            return return_obj

        # Build up the URL and header information.
        headers = {
            'Accept': 'application/json,text/html,application/xhtml+xml,application/xml'
        }
        self.logger.debug("Headers are: {0}".format(headers))

        files = list()

        for file2post in fileinfo:
            files.append((file2post['file_formfield'],
                          (file2post['filename'], open(file2post['filename'], 'rb'), file2post['mimetype'])))

        self._suppress_ssl_warning()

        try:
            r = requests.put(url=self._url(uri=uri), data=data, auth=(self.user.username, self.user.password),
                             files=files, verify=False, headers=headers)
            return_obj['changed'] = True  # POST of file would be a change
            self._process_response(return_obj=return_obj, http_response=r, ignore_error=ignore_error)

        except requests.exceptions.ConnectionError:
            if not ignore_error:
                self.logger.critical("Failed to connect to server.")
                raise IBMError("HTTP Return code: 502", "Failed to connect to server")
            else:
                self.logger.debug("Failed to connect to server.")
                return_obj.rc = 502

        return return_obj

    def invoke_get_file(self, description, uri, filename, no_headers=False, mime_types=None, ignore_error=False, requires_modules=None,
                        requires_version=None, warnings=[]):
        """
        Invoke a GET request and download the response data to a file
        """
        self._log_desc(description=description)

        warnings, return_call = self._process_warnings(uri=uri, requires_modules=requires_modules,
                                                       requires_version=requires_version,
                                                       warnings=warnings)
        return_obj = self.create_return_object(warnings=warnings)
        if return_call:
            return return_obj

        # In some cases passing a header causes response to come back as JSON
        if no_headers is True:
            headers = {}
        elif mime_types is None:
            headers = {
                'Accept': 'application/json,application/octet-stream'
            }
        else:
            headers = {
                'Accept': mime_types
            }
        self.logger.debug("Headers are: {0}".format(headers))

        self._suppress_ssl_warning()

        try:
            r = requests.get(url=self._url(uri=uri), auth=(self.user.username, self.user.password), verify=False,
                             stream=True, headers=headers, allow_redirects=False)

            if (r.status_code != 200 and r.status_code != 204 and r.status_code != 201):
                self.logger.error("  Request failed: ")
                self.logger.error("     status code: {0}".format(r.status_code))
                if r.text != "":
                    self.logger.error("     text: " + r.text)
                if not ignore_error:
                    raise IBMError("HTTP Return code: {0}".format(r.status_code), r.text)
                else:
                    return_obj['rc'] = r.status_code
                    return_obj['data'] = {'msg': 'Unable to extract contents to file!'}
            else:
                with open(filename, 'wb') as f:
                    for chunk in r.iter_content(chunk_size=1024):
                        if chunk:  # filter out keep-alive new chunks
                            f.write(chunk)
                return_obj['rc'] = 0
                return_obj['data'] = {'msg': 'Contents extracted to file: ' + filename}

        except requests.exceptions.ConnectionError:
            self._process_connection_error(ignore_error=ignore_error, return_obj=return_obj)

        except IOError:
            if not ignore_error:
                self.logger.critical("Failed to write to file: " + filename)
                raise IBMError("HTTP Return code: 999", "Failed to write to file: " + filename)
            else:
                self.logger.debug("Failed to write to file: " + filename)
                return_obj['rc'] = 999

        return return_obj

    def _invoke_request(self, func, description, uri, ignore_error, data={}, requires_modules=None,
                        requires_version=None, warnings=[]):
        """
        Send a request to the LMI.  This function is private and should not be
        used directly.  The invoke_get/invoke_put/etc functions should be used instead.
        """
        self._log_desc(description=description)

        warnings, return_call = self._process_warnings(uri=uri, requires_modules=requires_modules,
                                                       requires_version=requires_version,
                                                       warnings=warnings)
        return_obj = self.create_return_object(warnings=warnings)
        if return_call:
            return return_obj

        # There maybe some cases when header should be blank (not json)
        headers = {
            'Accept': 'application/json',
            'Content-type': 'application/json'
        }
        self.logger.debug("Headers are: {0}".format(headers))

        # Process the input data into JSON
        json_data = json.dumps(data)

        self.logger.debug("Input Data: " + json_data)

        self._suppress_ssl_warning()

        try:
            if func == requests.get or func == requests.delete:

                if data != {}:
                    r = func(url=self._url(uri), data=json_data, auth=(self.user.username, self.user.password),
                             verify=False, headers=headers, allow_redirects=False)
                else:
                    r = func(url=self._url(uri), auth=(self.user.username, self.user.password),
                             verify=False, headers=headers, allow_redirects=False)
            else:
                r = func(url=self._url(uri), data=json_data,
                         auth=(self.user.username, self.user.password),
                         verify=False, headers=headers)

            if func != requests.get:
                return_obj['changed'] = True  # Anything but GET should result in change

            self._process_response(return_obj=return_obj, http_response=r, ignore_error=ignore_error)

        except requests.exceptions.ConnectionError:
            self._process_connection_error(ignore_error=ignore_error, return_obj=return_obj)

        return return_obj

    def invoke_put(self, description, uri, data, ignore_error=False, requires_modules=None, requires_version=None,
                   warnings=[]):
        """
        Send a PUT request to the LMI.
        """
        return self._invoke_request(requests.put, description, uri, ignore_error, data,
                                    requires_modules=requires_modules, requires_version=requires_version,
                                    warnings=warnings)

    def invoke_post(self, description, uri, data, ignore_error=False, requires_modules=None, requires_version=None,
                    warnings=[]):
        """
        Send a POST request to the LMI.
        """
        return self._invoke_request(requests.post, description, uri, ignore_error, data,
                                    requires_modules=requires_modules, requires_version=requires_version,
                                    warnings=warnings)

    def invoke_get(self, description, uri, ignore_error=False, requires_modules=None, requires_version=None,
                   warnings=[]):
        """
        Send a GET request to the LMI.
        """
        return self._invoke_request(requests.get, description, uri, ignore_error, requires_modules=requires_modules,
                                    requires_version=requires_version, warnings=warnings)

    def invoke_delete(self, description, uri, ignore_error=False, requires_modules=None, requires_version=None,
                      warnings=[]):
        """
        Send a DELETE request to the LMI.
        """
        return self._invoke_request(requests.delete, description, uri, ignore_error, requires_modules=requires_modules,
                                    requires_version=requires_version, warnings=warnings)

    def get_facts(self):
        """
        Get facts about the appliance
        """
        # Fact collection will abort on any exception
        try:
            self.get_version()
        # Exceptions like those connection related will be ignored
        except:
            pass

    def get_version(self):
        """
        Get  appliance version (active partition)

        When firmware are installed or partition are changed, then this value is updated
        """
        self.facts['version'] = None
        import ibmsecurity.isvg.firmware

        ret_obj = ibmsecurity.isvg.firmware.get(self)
        for partition in ret_obj['data']:
            if partition['active'] is True:
                ver = partition['firmware_version'].split(' ')
                self.facts['version'] = ver[-1]



import logging
from ibmsecurity.utilities import tools

logger = logging.getLogger(__name__)

# URI for this module
uri = "/wga/kerberos/config/realms"
requires_modules = ['wga']
requires_version = None
import ibmsecurity.isam.web.kerberos_configuration.realms as realms
import ibmsecurity.isam.web.kerberos_configuration.realms_subsection as subsections


def search(isamAppliance, realm, propname, subsection=None, includeValuesInLine='yes', check_mode=False, force=False):
    """
    Search kerberos realm property by name
    """
    if subsection is None:
        uriprop = "{0}".format(realm)
        subsection = "NA"
    else:
        uriprop = "{0}/{1}".format(realm, subsection)

    logger.info("property to search under: {0} where property is: {1}".format(uriprop, propname))
    ret_obj = isamAppliance.invoke_get("Retrieve realm configuration", "{0}/{1}{2}".format(uri, realm,
                                                                                           tools.create_query_string(
                                                                                               includeValuesInLine=includeValuesInLine)),
                                       requires_modules=requires_modules, requires_version=requires_version)

    return_obj = isamAppliance.create_return_object()
    return_obj["warnings"] = ret_obj["warnings"]
    return_obj['data'] = {}

    for obj in ret_obj['data']:
        if obj['type'] == "property":
            if "{0} = ".format(propname) in obj['name']:
                logger.info("Found Kerberos realm property {0} id: {1}".format(propname, obj['id']))
                return_obj['data']['name'] = obj['name']
                return_obj['data']['id'] = obj['id']
                return_obj['rc'] = 0
                break

    return return_obj


def _check(isamAppliance, realm, propname, propvalue, subsection=None):
    """
    Check if kerberos realm's property with given value is already created or not

    :param isamAppliance:
    :return:
    """

    propstring = "{0} = {1}".format(propname, propvalue)

    ret_obj = search(isamAppliance, realm, propname, subsection)
    logger.debug("Looking for existing kerberos property {0} in {1}".format(propname, ret_obj['data']))
    if ret_obj['data'] != {}:
        if ret_obj['data']['name'] == propstring:
            logger.debug("Found kerberos property: {0} in : {1}".format(propname, ret_obj['data']))
            return True
    return False


def add(isamAppliance, realm, propname, propvalue, subsection=None, check_mode=False, force=False):
    """
            Add kerberos realm's property
            realm and subsection if passed needs to exist
            :param isamAppliance:
            :return:
         """
    if subsection is None:
        uriprop = realm
    else:
        uriprop = "{0}/{1}".format(realm, subsection)

    if realms.search(isamAppliance, realm) == {}:
        return isamAppliance.create_return_object(warnings=["Realm: {0} does not exists: ".format(realm)])

    if subsection is not None and subsections._check(isamAppliance, realm, subsection) is False:
        return isamAppliance.create_return_object(
            warnings=["Subsection: {0}/{1} does not exists.".format(realm, subsection)])

    if force is False and _check(isamAppliance, realm, propname, propvalue, subsection) is True:
        return isamAppliance.create_return_object(
            warnings=["Property: {0} with value: {1} already exists: ".format(uriprop, propvalue)])

    if force is True or _check(isamAppliance, realm, propname, propvalue, subsection) is False:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_post("Add kerberos real property", "{0}/{1}".format(uri, uriprop),
                                             {
                                                 "name": propname,
                                                 "value": propvalue
                                             },
                                             requires_modules=requires_modules, requires_version=requires_version)

    return isamAppliance.create_return_object()


def delete(isamAppliance, realm, propname, subsection=None, check_mode=False, force=False):
    """
    Remove kerberos realms property
    :param isamAppliance:
    :return:
    """
    if realms.search(isamAppliance, realm) == {}:
        return isamAppliance.create_return_object(warnings=["Realm: {0} does not exists: ".format(realm)])

    if subsection is None:
        uriprop = realm
    else:
        uriprop = "{0}/{1}".format(realm, subsection)

    logger.debug(" Remove param uri = {0}/{1}/{2}".format(uri, uriprop, propname))
    ret_obj = search(isamAppliance, realm, propname, subsection)

    if ret_obj['data'] == {} and force is False:
        return isamAppliance.create_return_object(
            warnings=["property: {0} not found or force is: {1}".format(propname, force)])
    else:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_delete(description="Delete kerberos realm prop ",
                                               uri="{0}/{1}/{2}".format(uri, uriprop, propname),
                                               requires_modules=requires_modules, requires_version=requires_version)


def update(isamAppliance, realm, propname, propvalue, subsection=None, check_mode=False, force=False):
    """
    Set kerberos realm property

    :param isamAppliance:
    :return:
    """
    if subsection is None:
        uriprop = realm
    else:
        uriprop = "{0}/{1}".format(realm, subsection)

    if realms.search(isamAppliance, realm) == {}:
        return isamAppliance.create_return_object(warnings=["Realm: {0} does not exists: ".format(realm)])

    if subsection is not None and subsections._check(isamAppliance, realm, subsection) is False:
        return isamAppliance.create_return_object(
            warnings=["Subsection: {0}/{1} does not exists.".format(realm, subsection)])

    ret_obj = search(isamAppliance, realm, propname, subsection)
    warnings = ret_obj["warnings"]

    if ret_obj["data"] == {}:
        warnings.append("Property {0} not found, skipping update.".format(propname))
        return isamAppliance.create_return_object(warnings=warnings)

    needs_update = False

    propstring = "{0} = {1}".format(propname, propvalue)

    if force is not True:
        if ret_obj['data']['name'] != propstring:
            needs_update = True

    if force is True or needs_update is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True, warnings=warnings)
        else:
            idval = "/{0}/{1}".format(uriprop, propname)
            return isamAppliance.invoke_put(description="Add kerberos param value",
                                            uri="{0}/{1}/{2}".format(uri, uriprop, propname),
                                            data={
                                                "id": idval,
                                                "value": propvalue
                                            },
                                            requires_modules=requires_modules, requires_version=requires_version,
                                            warnings=warnings)
    return isamAppliance.create_return_object(warnings=warnings)


def set(isamAppliance, realm, propname, propvalue, subsection=None, check_mode=False, force=False):
    """
    Set kerberos realm property

    :param isamAppliance:
    :return:
    """
    if realms.search(isamAppliance, realm) == {}:
        return isamAppliance.create_return_object(warnings=["Realm: {0} does not exists: ".format(realm)])

    if subsection is not None and subsections._check(isamAppliance, realm, subsection) is False:
        return isamAppliance.create_return_object(
            warnings=["Subsection: {0}/{1} does not exists.".format(realm, subsection)])

    ret_obj = search(isamAppliance, realm, propname, subsection)

    if ret_obj["data"] == {}:
        return add(isamAppliance, realm, propname, propvalue, subsection, check_mode, force=True)

    return update(isamAppliance, realm, propname, propvalue, subsection, check_mode, force)



import logging
from ibmsecurity.utilities import tools

logger = logging.getLogger(__name__)

uri = "/isam/dsc/config"
requires_modules = None
requires_version = None
requires_model = "Docker"


def get(isamAppliance, check_mode=False, force=False):
    """
    Retrieve the current distributed session cache policy
    """
    return isamAppliance.invoke_get("Retrieve the current distributed session cache policy", uri,
                                    requires_modules=requires_modules, requires_version=requires_version,
                                    requires_model=requires_model)


def set(isamAppliance, service_port=443, replication_port=444, worker_threads=64, max_session_lifetime=3600,
        client_grace=600, servers=[], check_mode=False, force=False):
    """
    Update the current distributed session cache policy
    """
    # Create a simple json with just the main client attributes
    dsc_json = {
        "worker_threads": worker_threads,
        "max_session_lifetime": max_session_lifetime,
        "client_grace": client_grace,
        "service_port": service_port,
        "replication_port": replication_port,
        "servers": servers
    }

    obj = _check(isamAppliance, dsc_json)
    if force is True or obj['value'] is False:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True, warnings=obj['warnings'])
        else:
            return isamAppliance.invoke_put("Update the current distributed session cache policy", uri, dsc_json,
                                            requires_modules=requires_modules, requires_version=requires_version,
                                            requires_model=requires_model, warnings=obj['warnings'])

    return isamAppliance.create_return_object(warnings=obj['warnings'])


def _check(isamAppliance, cluster_json):
    """
    Check if provided json values match the configuration on appliance

    :param isamAppliance:
    :param cluster_json:
    :return:
    """

    obj = {"value": False, "warnings": ""}

    ret_obj = get(isamAppliance)
    sorted_ret_obj = tools.json_sort(ret_obj['data'])
    sorted_json_data = tools.json_sort(cluster_json)
    logger.debug("Sorted Existing Data:{0}".format(sorted_ret_obj))
    logger.debug("Sorted Desired  Data:{0}".format(sorted_json_data))

    if sorted_ret_obj != sorted_json_data:
        logger.info("Existing and input data do not match - updated needed.")
        obj['value'] = False
        obj['warnings'] = ret_obj['warnings']
        return obj
    else:
        obj['value'] = True
        obj['warnings'] = ret_obj['warnings']
        return obj


def compare(isamAppliance1, isamAppliance2):
    """
    Compare DSC configuration between two appliances
    """

    ret_obj1 = get(isamAppliance1)
    ret_obj2 = get(isamAppliance2)

    return tools.json_compare(ret_obj1, ret_obj2, deleted_keys=[])



import logging

from ibmsecurity.utilities import tools

logger = logging.getLogger(__name__)

uri = "/wga/apiac/credentials"
requires_modules = ["wga"]
requires_version = "9.0.7"


def get(isamAppliance, check_mode=False, force=False):
    """
    Retrieve the stored ISAM credential
    """
    return isamAppliance.invoke_get("Retrieve the stored ISAM credential",
                                    "{0}".format(uri),
                                    requires_modules=requires_modules, requires_version=requires_version)


def add(isamAppliance, admin_id, admin_pwd, admin_domain="Default", check_mode=False, force=False):
    """
    Store the ISAM administrator credentials
    """
    exist, warnings = _check(isamAppliance)

    if force is True or exist is False:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True, warnings=warnings)
        else:
            return isamAppliance.invoke_post("Store the ISAM administrator credentials",
                                         "{0}".format(uri),
                                         {
                                             'admin_id': admin_id,
                                             'admin_pwd': admin_pwd,
                                             'admin_domain': admin_domain
                                         },
                                         requires_modules=requires_modules, requires_version=requires_version)

    return isamAppliance.create_return_object(warnings=warnings)


def delete(isamAppliance, check_mode=False, force=False):
    """
    Delete the stored ISAM administrator credential
    """
    exist, warnings = _check(isamAppliance)

    if force is True or exist is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True, warnings=warnings)
        else:
            return isamAppliance.invoke_delete("Delete the stored ISAM administrator credential",
                                               "{0}".format(uri),
                                               requires_modules=requires_modules, requires_version=requires_version)

    return isamAppliance.create_return_object(warnings=warnings)


def _check(isamAppliance, check_mode=False, force=False):
    ret_obj = get(isamAppliance)
    if ret_obj['data'] == {}:
        return False, ret_obj['warnings']
    elif ret_obj['data']['admin_id'] is None:
        return False, ret_obj['warnings']
    else:
        return True, ret_obj['warnings']



import logging

logger = logging.getLogger(__name__)

# URI for this module
uri = "/mga/connector_instructions"
requires_modules = ["federation"]
requires_version = "9.0.0.0"


def get_all(isamAppliance, check_mode=False, force=False):
    """
    Retrieve all connector instruction sets
    """
    return isamAppliance.invoke_get("Retrieve all connector instruction sets", uri,
                                    requires_modules=requires_modules,
                                    requires_version=requires_version)


def get(isamAppliance, id, variable_names=None, check_mode=False, force=False):
    """
    Retrieve a connector instruction set

    Noe: Pass variable_names as a list key value pairs like so:
        [{'hostname': 'ibm.com'}, {'key': 'value'}]

    TODO: test variable_names...
    """
    query_str = ""
    if variable_names is not None:
        for var in variable_names:
            if query_str == "":
                query_str = '?'
            else:
                query_str += '&'

            for key, value in var.items():
                query_str += "{0}={1}".format(key, value)

    return isamAppliance.invoke_get("Retrieve a connector instruction set",
                                    "{0}/{1}{2}".format(uri, id, query_str),
                                    requires_modules=requires_modules,
                                    requires_version=requires_version)



import logging

logger = logging.getLogger(__name__)

# URI for this module
uri = "/core/net/connect"
requires_modules = None
requires_version = "9.0.3.0"
requires_model = "Appliance"


def connect(isamAppliance, server, port, ssl=True, timeout=60, key=None, showcerts=None, status=None, reconnect=None,
            pause=None, debug=None, msg=None, nbio_test=None, state=None, nbio=None, crlf=None, quiet=None,
            ssl2=None, ssl3=None, tls1_2=None, tls1_1=None, tls1=None, dtls1=None, no_ssl2=None, no_ssl3=None,
            no_tls1_2=None, no_tls1_1=None, no_tls1=None, tlsextdebug=None, check_mode=False, force=False):
    """
    Run Connect Test
    """
    # Create JSON with required fields
    json_data = {
        'server': server,
        'port': port
    }

    # Populate JSON with optional values that are specified
    default_data = {
        'ssl': ssl,
        'timeout': timeout,
        'key': key,
        'showcerts': showcerts,
        'status': status,
        'reconnect': reconnect,
        'pause': pause,
        'debug': debug,
        'msg': msg,
        'nbio_test': nbio_test,
        'state': state,
        'nbio': nbio,
        'crlf': crlf,
        'quiet': quiet,
        'ssl2': ssl2,
        'ssl3': ssl3,
        'tls1_2': tls1_2,
        'tls1_1': tls1_1,
        'tls1': tls1,
        'dtls1': dtls1,
        'no_ssl2': no_ssl2,
        'no_ssl3': no_ssl3,
        'no_tls1_2': no_tls1_2,
        'no_tls1_1': no_tls1_1,
        'no_tls1': no_tls1,
        'tlsextdebug': tlsextdebug
    }

    for k, value in default_data.items():
        if value is not None:
            json_data[k] = value

    ret_obj = isamAppliance.invoke_post("Run Connect Test", uri, json_data, requires_modules=requires_modules,
                                        requires_version=requires_version, requires_model=requires_model,
                                        ignore_error=True)
    # HTTP POST calls get flagged as changes - but test connection changes nothing so override
    if ret_obj['changed'] is True:
        ret_obj['changed'] = False

    return ret_obj



import logging

logger = logging.getLogger(__name__)

requires_model="Appliance"

def export(isamAppliance, filename, check_mode=False, force=False):
    """
    Retrieve the cluster signature file
    """
    import os.path

    if force is True or os.path.exists(filename) is False:
        if check_mode is False:  # No point downloading a file if in check_mode
            return isamAppliance.invoke_get_file(
                "Retrieve the cluster signature file",
                "/isam/cluster/signature/v1",
                filename, requires_model=requires_model)

    return isamAppliance.create_return_object()



import logging
from io import open
import ibmsecurity.utilities.tools

logger = logging.getLogger(__name__)

# URI for this module
uri = "/license_activate"
requires_modules = None
requires_version = None


def get(isdsAppliance, check_mode=False, force=False):
    """
    Retrieve license installed in the appliance
    """
    return isdsAppliance.invoke_get("Retrieve license installed in the appliance",
                                    "{0}".format(uri), requires_modules=requires_modules,
                                    requires_version=requires_version)


def install(isdsAppliance, license, check_mode=False, force=False):
    """
    install license in the appliance
    license should be the filename of the license file
    license file content expected format:

    IBM Security Directory Suite V8.0.x
    xxx Edition Key:     xyxyxyxyxyxyxyxyxyxyxy
    """
    f = open(license, 'rt')
    logger.debug(f.readline())
    licence_key_line = f.readline()

    licence_key_tokens = licence_key_line.split(':')
    license_name = licence_key_tokens[0].strip()
    license_key = licence_key_tokens[1].strip()
    logger.debug("License name: " + license_name)
    logger.debug("License key: " + license_key)

    if _check_license(isdsAppliance) == True and force == False:
        return isdsAppliance.create_return_object(warnings=["License already installed"])
    else:
        if _check_license(isdsAppliance) == False or force == True:
            if check_mode:
                return isdsAppliance.create_return_object(
                    changed=True)  # there is no real change in appliance, so maybe it should be false
            else:
                return isdsAppliance.invoke_post("Applying license key to appliance", uri,
                                                 {
                                                     "Licensetext": license_key
                                                 })


def _check_license(isdsAppliance):
    """
    check if a particular license is installed in the appliance
    cannot retrieve installed license key value so sole presence
    of license is sufficient
    """

    ret_obj = get(isdsAppliance)
    """
    Returned json is any of the following:
        {"LicenseActivated":"LIMITED"}
        {"LicenseActivated":"STANDARD"}
        {"LicenseActivated":"ENTERPRISE"}
    """

    if ret_obj['data'] == {"LicenseActivated": "STANDARD"}:
        return True
    elif ret_obj['data'] == {"LicenseActivated": "ENTERPRISE"}:
        return True
    else:
        return False  # no license installed in the appliance


def compare(isdsAppliance1, isdsAppliance2):
    """
    Compare license installed between two appliances
    """
    ret_obj1 = get(isdsAppliance1)
    ret_obj2 = get(isdsAppliance2)

    import ibmsecurity.utilities.tools
    return ibmsecurity.utilities.tools.json_compare(ret_obj1, ret_obj2, deleted_keys=['serial_number'])



import logging
from ibmsecurity.utilities import tools

logger = logging.getLogger(__name__)

# URI for this module
uri = "/iam/access/v8/policysets"


def get_all(isamAppliance, filter=None, sortBy=None, check_mode=False, force=False):
    """
    Retrieve a list of policy sets
    """
    return isamAppliance.invoke_get("Retrieve a list of policy sets",
                                    "{0}/{1}".format(uri, tools.create_query_string(filter=filter, sortBy=sortBy)))


def get(isamAppliance, name, check_mode=False, force=False):
    """
    Retrieve a specific policy set
    """
    ret_obj = search(isamAppliance, name=name, check_mode=check_mode, force=force)
    pol_id = ret_obj['data']

    if pol_id == {}:
        logger.info("Policy {0} had no match, skipping retrieval.".format(name))
        return isamAppliance.create_return_object()
    else:
        return isamAppliance.invoke_get("Retrieve a specific policy set",
                                        "{0}/{1}".format(uri, pol_id))


def get_policies(isamAppliance, name, check_mode=False, force=False):
    """
    Retrieve policies in a specific policy set
    """
    ret_obj = search(isamAppliance, name=name, check_mode=check_mode, force=force)
    pol_id = ret_obj['data']

    if pol_id == {}:
        logger.info("Policy {0} had no match, skipping retrieval.".format(name))
        return isamAppliance.create_return_object()
    else:
        return isamAppliance.invoke_get("Retrieve policies in a specific policy set",
                                        "{0}/{1}/policies".format(uri, pol_id))


def search(isamAppliance, name, force=False, check_mode=False):
    """
    Search policy set id by name
    """
    ret_obj = get_all(isamAppliance)
    return_obj = isamAppliance.create_return_object()

    for obj in ret_obj['data']:
        if obj['name'] == name:
            logger.info("Found Policy Set {0} id: {1}".format(name, obj['id']))
            return_obj['data'] = obj['id']
            return_obj['rc'] = 0

    return return_obj


def set(isamAppliance, name, policies=None, policyCombiningAlgorithm='denyOverrides', description="", predefined=False,
        new_name=None, check_mode=False, force=False):
    """
    Creating or Modifying a Policy Set

    Note: Please input policies as an array of policy names (it will be converted to id's)
    """
    if (search(isamAppliance, name=name))['data'] == {}:
        # Force the add - we already know policy set does not exist
        logger.info("Policy Set {0} had no match, requesting to add new one.".format(name))
        return add(isamAppliance, name, policies, policyCombiningAlgorithm, description, predefined, check_mode, True)
    else:
        # Update request
        logger.info("Policy Set {0} exists, requesting to update.".format(name))
        return update(isamAppliance, name, policies, policyCombiningAlgorithm, description, predefined, new_name,
                      check_mode, force)


def add(isamAppliance, name, policies=None, policyCombiningAlgorithm='denyOverrides', description="", predefined=False,
        check_mode=False, force=False):
    """
    Create a new Policy Set

    Note: Please input policies as an array of policy names (it will be converted to id's)
    """
    if force is False:
        ret_obj = search(isamAppliance, name)

    if force is True or ret_obj['data'] == {}:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            json_data = {
                "name": name,
                "description": description,
                "predefined": predefined
            }
            if policies is not None:
                json_data['policies'] = _convert_policy_name_to_id(isamAppliance, policies)
            if policyCombiningAlgorithm is not None:
                json_data['policyCombiningAlgorithm'] = policyCombiningAlgorithm
            return isamAppliance.invoke_post(
                "Create a new Policy Set", uri, json_data)

    return isamAppliance.create_return_object()


def _convert_policy_name_to_id(isamAppliance, policies):
    pol_ids = []
    import ibmsecurity.isam.aac.access_control.policies
    for pol_name in policies:
        ret_obj = ibmsecurity.isam.aac.access_control.policies.search(isamAppliance, pol_name)
        pol_id = ret_obj['data']
        if pol_id != {}:
            pol_ids.append(pol_id)
            logger.debug("Converting policy {0} to ID: {1}".format(pol_name, pol_id))
        else:
            logger.warning("Unable to find policy {0}, skipping.".format(pol_name))

    return pol_ids


def _convert_policy_id_to_name(isamAppliance, policies):
    pol_names = []
    import ibmsecurity.isam.aac.access_control.policies
    for pol_id in policies:
        ret_obj = ibmsecurity.isam.aac.access_control.policies._get(isamAppliance, pol_id)
        if ret_obj['data'] != {}:
            pol_name = ret_obj['data']['name']
            pol_names.append(pol_name)
            logger.debug("Converting policy {0} to Name: {1}".format(pol_id, pol_name))
        else:
            logger.warning("Unable to find policy {0}, skipping.".format(pol_id))

    return pol_names


def delete(isamAppliance, name, check_mode=False, force=False):
    """
    Delete a Policy Set
    """
    ret_obj = search(isamAppliance, name, check_mode=check_mode, force=force)
    pol_id = ret_obj['data']

    if pol_id == {}:
        logger.info("Policy Set {0} not found, skipping delete.".format(name))
    else:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_delete(
                "Delete a Policy Set",
                "{0}/{1}".format(uri, pol_id))

    return isamAppliance.create_return_object()


def update(isamAppliance, name, policies=None, policyCombiningAlgorithm='denyOverrides', description="",
           predefined=False,
           new_name=None, check_mode=False, force=False):
    """
    Update a specified policy set

    Note: Please input policies as an array of policy names (it will be converted to id's)
    """
    pol_id, update_required, json_data = _check(isamAppliance, name, policies, policyCombiningAlgorithm, description,
                                                predefined, new_name)
    if pol_id is None:
        from ibmsecurity.appliance.ibmappliance import IBMError
        raise IBMError("999", "Cannot update data for unknown policy set: {0}".format(name))

    if force is True or update_required is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_put(
                "Update a specified policy set",
                "{0}/{1}".format(uri, pol_id), json_data)

    return isamAppliance.create_return_object()


def _check(isamAppliance, name, policies, policyCombiningAlgorithm, description, predefined, new_name):
    """
    Check and return True if update needed
    """
    update_required = False
    json_data = {
        "name": name,
        "description": description,
        "predefined": predefined
    }
    ret_obj = get(isamAppliance, name)
    if ret_obj['data'] == {}:
        logger.warning("Policy Set not found, returning no update required.")
        return None, update_required, json_data
    else:
        pol_id = ret_obj['data']['id']
        if new_name is not None:
            json_data['name'] = new_name
        else:
            json_data['name'] = name
        if policies is not None:
            json_data['policies'] = _convert_policy_name_to_id(isamAppliance, policies)
        else:
            try:
                del ret_obj['data']['policies']
            except:
                pass
        if policyCombiningAlgorithm is not None:
            json_data['policyCombiningAlgorithm'] = policyCombiningAlgorithm
        else:
            try:
                del ret_obj['data']['policyCombiningAlgorithm']
            except:
                pass
        del ret_obj['data']['id']
        del ret_obj['data']['userlastmodified']
        del ret_obj['data']['lastmodified']
        del ret_obj['data']['datecreated']
        # TODO Check if sorting of policies is a valid for comparison (order maybe important!)
        import ibmsecurity.utilities.tools
        sorted_json_data = ibmsecurity.utilities.tools.json_sort(json_data)
        logger.debug("Sorted input: {0}".format(sorted_json_data))
        sorted_ret_obj = ibmsecurity.utilities.tools.json_sort(ret_obj['data'])
        logger.debug("Sorted existing data: {0}".format(sorted_ret_obj))
        if sorted_ret_obj != sorted_json_data:
            logger.info("Changes detected, update needed.")
            update_required = True

    return pol_id, update_required, json_data


def update_policies(isamAppliance, name, policies, action, check_mode=False, force=False):
    """
    Update a specified policy set's policies (add/remove/set)

    Note: Please input policies as an array of policy names (it will be converted to id's)
    """
    pol_id, update_required, json_data = _check_policies(isamAppliance, name, policies, action)

    if pol_id is None:
        from ibmsecurity.appliance.ibmappliance import IBMError
        raise IBMError("999", "Cannot update data for unknown policy set: {0}".format(name))

    if force is True or update_required is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_put(
                "Update a specified policy set",
                "{0}/{1}/policies{2}".format(uri, pol_id, tools.create_query_string(action=action)), json_data)

    return isamAppliance.create_return_object()


def _check_policies(isamAppliance, name, policies, action):
    update_required = False
    json_data = {}
    pol_ids = []
    ret_obj = get(isamAppliance, name)
    if ret_obj['data'] == {}:
        logger.warning("Policy Set not found, returning no update required.")
        return None, update_required, json_data
    else:
        pol_id = ret_obj['data']['id']
        new_pol_ids = _convert_policy_name_to_id(isamAppliance, policies)
        if action == 'set':
            pol_ids = new_pol_ids
        else:
            for new_pol_id in new_pol_ids:
                if new_pol_id in ret_obj['data']['policies']:
                    exists = True
                else:
                    exists = False
                if action == 'add':
                    if exists:
                        logger.info("Policy ID {0} already exists skipping.".format(new_pol_id))
                    else:
                        pol_ids.append(new_pol_ids)
                        logger.info("Policy ID {0} does not exist, appending to list for additon.".format(new_pol_id))
                elif action == 'delete':
                    if exists:
                        pol_ids.append(new_pol_ids)
                        logger.info("Policy ID {0} exists, appending to list for deletion.".format(new_pol_id))
                    else:
                        logger.info("Policy ID {0} does not exist skipping.")
                else:
                    from ibmsecurity.appliance.ibmappliance import IBMError
                    raise IBMError("999", "Unknown action: {0}.".format(action))

        json_data['policies'] = pol_ids
        return pol_id, update_required, json_data


def compare(isamAppliance1, isamAppliance2):
    """
    Compare Policy Sets between two appliances
    """
    ret_obj1 = get_all(isamAppliance1)
    ret_obj2 = get_all(isamAppliance2)

    for obj in ret_obj1['data']:
        del obj['id']
        del obj['userlastmodified']
        del obj['lastmodified']
        del obj['datecreated']
        obj['policies'] = _convert_policy_id_to_name(isamAppliance1, obj['policies'])
    for obj in ret_obj2['data']:
        del obj['id']
        del obj['userlastmodified']
        del obj['lastmodified']
        del obj['datecreated']
        obj['policies'] = _convert_policy_id_to_name(isamAppliance2, obj['policies'])

    return tools.json_compare(ret_obj1, ret_obj2,
                              deleted_keys=['id', 'userlastmodified', 'lastmodified', 'datecreated'])



import logging
from ibmsecurity.utilities import tools

logger = logging.getLogger(__name__)

# URI for this module
uri = "/wga/kerberos/config/libdefaults"
requires_modules = ['wga']
requires_version = None


def get_all(isamAppliance, recursive='yes', includeValues='yes', check_mode=False, force=False):
    """
    Retrieve Kerberos Configuration: Defaults
    """
    return isamAppliance.invoke_get("Retrieve Kerberos Configuration: Defaults",
                                    "{0}{1}".format(uri, tools.create_query_string(recursive=recursive,
                                                                                   includeValues=includeValues)),
                                    requires_modules=requires_modules, requires_version=requires_version)


def get(isamAppliance, name, check_mode=False, force=False):
    """
    Retrieve a specific Kerberos Configuration Default Property
    """
    ret_obj = search(isamAppliance, name=name, check_mode=check_mode, force=force)
    config_prop_name = ret_obj['data']
    warnings = ret_obj["warnings"]

    if config_prop_name == {}:
        logger.info("Kerberos Default config property {0} had no match, skipping retrieval.".format(name))
        warnings.append("Kerberos Default config property {0} had no match.".format(name))
        return isamAppliance.create_return_object(warnings=warnings)
    else:
        return _get(isamAppliance, name)


def _get(isamAppliance, config_prop_name, recursive="yes", includeValues="yes"):
    return isamAppliance.invoke_get("Retrieve a specific Kerberos Configuration default property",
                                    "{0}/{1}".format(uri, config_prop_name), requires_modules=requires_modules,
                                    requires_version=requires_version)


def search(isamAppliance, name, check_mode=False, force=False):
    """
    Search default kerberos config by name
    """
    ret_obj = get_all(isamAppliance)
    return_obj = isamAppliance.create_return_object()
    return_obj["warnings"] = ret_obj["warnings"]

    for obj in ret_obj['data']:
        if obj['name'] == name:
            logger.info("Found Kerberos Config Default property {0} id: {1}".format(name, obj['id']))
            return_obj['data'] = obj['id']
            return_obj['rc'] = 0

    return return_obj


def add(isamAppliance, name, value, check_mode=False, force=False):
    """
    Create Kerberos Configuration Default Property
    """
    ret_obj = search(isamAppliance, name=name, check_mode=check_mode, force=force)
    warnings = ret_obj["warnings"]
    if force is True or ret_obj["data"] == {}:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True, warnings=warnings)
        else:
            return isamAppliance.invoke_post("Create Kerberos Configuration Default Property", uri,
                                             {
                                                 "name": name,
                                                 "value": value
                                             },
                                             requires_modules=requires_modules, requires_version=requires_version,
                                             warnings=warnings)

    return isamAppliance.create_return_object(warnings=warnings)


def delete(isamAppliance, name, check_mode=False, force=False):
    """
    Delete an Kerberos Configuration Default property
    """
    ret_obj = search(isamAppliance, name, check_mode=check_mode, force=force)
    config_prop_name = ret_obj['data']
    warnings = ret_obj["warnings"]

    if config_prop_name == {}:
        logger.info("Default property {0} not found, skipping delete.".format(name))
    else:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True, warnings=warnings)
        else:
            return isamAppliance.invoke_delete(
                "Delete an Kerberos configuration default property",
                "{0}/{1}".format(uri, name), requires_modules=requires_modules,
                requires_version=requires_version, warnings=warnings)

    return isamAppliance.create_return_object(warnings=warnings)


def update(isamAppliance, name, value, check_mode=False,
           force=False):
    """
    Update a specified Kerberos Configuration Default property
    """
    ret_obj = get(isamAppliance, name)
    warnings = ret_obj["warnings"]

    if ret_obj["data"] == {}:
        warnings.append("Definiton {0} not found, skipping update.".format(name))
        return isamAppliance.create_return_object(warnings=warnings)
    else:
        config_prop_value = ret_obj["data"]
        logger.debug("property value is: {0}".format(config_prop_value))

    needs_update = False
    json_data = {
        "value": value
    }

    if force is not True:
        logger.debug(str(value))
        logger.debug(' '.join(config_prop_value))
        if str(value).lower() != (' '.join(config_prop_value)).lower():
            needs_update = True

    logger.debug("needs update is: {0}".format(needs_update))

    if force is True or needs_update is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True, warnings=warnings)
        else:
            return isamAppliance.invoke_put(
                "Update a specified Kerberos Configuration Default property",
                "{0}/{1}".format(uri, name), json_data, requires_modules=requires_modules,
                requires_version=requires_version, warnings=warnings)

    return isamAppliance.create_return_object(warnings=warnings)

def set(isamAppliance, name, value, check_mode=False, force=False):
    """
    Set specified Kerberos Configuration Default property

    :param isamAppliance:
    :return:
    """
    ret_obj = search(isamAppliance, name)

    if ret_obj["data"] == {}:
        return add(isamAppliance, name, value, check_mode, force=True)

    return update(isamAppliance,  name, value, check_mode, force)

def compare(isamAppliance1, isamAppliance2):
    """
    Compare kerberos configuration defaults between two appliances
    """
    ret_obj1 = get_all(isamAppliance1)
    ret_obj2 = get_all(isamAppliance2)

    for obj in ret_obj1['data']:
        del obj['id']
    for obj in ret_obj2['data']:
        del obj['id']

    return tools.json_compare(ret_obj1, ret_obj2, deleted_keys=['id'])



import logging
import os.path
import ibmsecurity.utilities.tools

logger = logging.getLogger(__name__)


def get_all(isamAppliance, kdb_id, check_mode=False, force=False):
    """
    Retrieving certificate request names and details in a certificate database
    """
    return isamAppliance.invoke_get("Retrieving certificate request names and details in a certificate database",
                                    "/isam/ssl_certificates/{0}/cert_request".format(kdb_id))


def get(isamAppliance, kdb_id, cert_id, check_mode=False, force=False):
    """
    Retrieving a certificate request from a certificate database
    """
    return isamAppliance.invoke_get("Retrieving a certificate request from a certificate database",
                                    "/isam/ssl_certificates/{0}/cert_request/{1}".format(kdb_id, cert_id))


def add(isamAppliance, kdb_id, label, dn, size='1024', signature_algorithm='', check_mode=False, force=False):
    """
    Creating a certificate request in a certificate database
    """
    
    warnings = []
    
    if signature_algorithm is not None:
        if ibmsecurity.utilities.tools.version_compare(isamAppliance.facts["version"], "9.0.2.0") < 0:
            warnings.append("Appliance at version: {0}, signature_algorithm is not supported. Needs 9.0.2.0 or higher. "
                            "Ignoring signature_algorithm for this call".format(isamAppliance.facts["version"]))
            json_obj = {
                    "label": label,
                    "dn": dn,
                    "size": size
            }
        else:
            json_obj = {
                    "label": label,
                    "dn": dn,
                    "size": size,
                    "signature_algorithm": signature_algorithm
            }
    
    if force is True or _check(isamAppliance, kdb_id, label) is False:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            
            return isamAppliance.invoke_post(
                "Creating a certificate request in a certificate database",
                "/isam/ssl_certificates/{0}/cert_request".format(kdb_id),
                json_obj,
                warnings=warnings)

    return isamAppliance.create_return_object(changed=False, warnings=warnings)


def delete(isamAppliance, kdb_id, cert_id, check_mode=False, force=False):
    """
    Deleting a certificate request from a certificate database
    """
    if force is True or _check(isamAppliance, kdb_id, cert_id) is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_delete(
                "Deleting a certificate request from a certificate database",
                "/isam/ssl_certificates/{0}/cert_request/{1}".format(kdb_id, cert_id))

    return isamAppliance.create_return_object()


def export_cert(isamAppliance, kdb_id, cert_id, filename, check_mode=False, force=False):
    """
    Exporting a certificate request from a certificate database
    """
    if force is True or (_check(isamAppliance, kdb_id, cert_id) is True and os.path.exists(filename) is False):
        if check_mode is False:  # No point downloading a file if in check_mode
            return isamAppliance.invoke_get_file(
                "Exporting a certificate request from a certificate database",
                "/isam/ssl_certificates/{0}/cert_request/{1}?export".format(kdb_id, cert_id),
                filename)

    return isamAppliance.create_return_object()


def _check(isamAppliance, kdb_id, cert_id):
    """
    Check if personal certificate already exists in certificate database
    """
    ret_obj = get_all(isamAppliance, kdb_id)

    for certdb in ret_obj['data']:
        if certdb['id'] == cert_id:
            return True

    return False


def compare(isamAppliance1, isamAppliance2, kdb_id):
    """
    Compare signer certificates in certificate database between two appliances
    """
    ret_obj1 = get_all(isamAppliance1, kdb_id)
    ret_obj2 = get_all(isamAppliance2, kdb_id)

    import ibmsecurity.utilities.tools
    return ibmsecurity.utilities.tools.json_compare(ret_obj1, ret_obj2, deleted_keys=[])



import logging
import ibmsecurity.utilities.tools
import os.path

logger = logging.getLogger(__name__)


def get_all(isamAppliance, check_mode=False, force=False):
    """
    Get information on existing groups
    """
    return isamAppliance.invoke_get("Retrieving groups", "/sysaccount/groups/v1")


def get(isamAppliance, id, check_mode=False, force=False):
    """
    Get information on particular group by id
    """
    return isamAppliance.invoke_get("Retrieving group", "/sysaccount/groups/{0}/v1".format(id))


def create(isamAppliance, id, check_mode=False, force=False):
    """
    Create a new group
    """
    if force is True or _check(isamAppliance, id=id) is False:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_post("Creating group", "/sysaccount/groups/v1",
                                             {
                                                 'id': id
                                             })

    return isamAppliance.create_return_object()


def _check(isamAppliance, id=None):
    """
    Check if the last created group has the exact same id or id exists

    :param isamAppliance:
    :param comment:
    :return:
    """
    ret_obj = get_all(isamAppliance)

    if id != None:
        for groups in ret_obj['data']:
            if groups['id'] == id:
                return True

    return False


def delete(isamAppliance, id, check_mode=False, force=False):
    """
    Delete a group
    """
    if force is True or _check(isamAppliance, id=id) is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_delete("Deleting group", "/sysaccount/groups/{0}/v1".format(id))

    return isamAppliance.create_return_object()


def compare(isamAppliance1, isamAppliance2):
    """
    Compare the list of groups between two appliances
    """
    ret_obj1 = get_all(isamAppliance1)
    ret_obj2 = get_all(isamAppliance2)

    return ibmsecurity.utilities.tools.json_compare(ret_obj1, ret_obj2)



import logging
from ibmsecurity.utilities import tools

logger = logging.getLogger(__name__)

# "uri" variable already used so using a different name
module_uri = "/iam/access/v8/authentication/policies"
requires_modules = ["mga"]
requires_version = None


def get_all(isamAppliance, start=None, count=None, filter=None, sortBy=None, check_mode=False, force=False):
    """
    Retrieve a list of authentication policies
    """
    return isamAppliance.invoke_get("Retrieve a list of authentication policies", "{0}/{1}".format(module_uri,
                                                                                                   tools.create_query_string(
                                                                                                       start=start,
                                                                                                       count=count,
                                                                                                       filter=filter,
                                                                                                       sortBy=sortBy)),
                                    requires_modules=requires_modules, requires_version=requires_version)


def get(isamAppliance, name, check_mode=False, force=False):
    """
    Retrieve a specific authentication policy
    """
    ret_obj = search(isamAppliance, name)
    if ret_obj['data'] != {}:
        return _get(isamAppliance, ret_obj['data'])
    else:
        return isamAppliance.create_return_object()


def _get(isamAppliance, id):
    """
    Retrieve a specific authentication policy
    """
    return isamAppliance.invoke_get("Retrieve a specific authentication policy",
                                    "{0}/{1}".format(module_uri, id), requires_modules=requires_modules,
                                    requires_version=requires_version)


def set_file(isamAppliance, name, policy_file, uri, description="",
             dialect="urn:ibm:security:authentication:policy:1.0:schema", enabled=None, check_mode=False, force=False):
    # Read policy from file and call set()
    with open(policy_file, 'r') as myfile:
        policy = myfile.read().replace('\n', '')
    return set(isamAppliance, name, policy, uri, description, dialect, enabled, check_mode, force)


def set(isamAppliance, name, policy, uri, description="", dialect="urn:ibm:security:authentication:policy:1.0:schema",
        enabled=None, check_mode=False, force=False):
    ret_obj = search(isamAppliance, name)
    if ret_obj['data'] == {}:
        return add(isamAppliance, name, policy, uri, description, dialect, enabled, check_mode, True)
    else:
        return update(isamAppliance, name, policy, uri, description, dialect, enabled, check_mode, force)


def add(isamAppliance, name, policy, uri, description="", dialect="urn:ibm:security:authentication:policy:1.0:schema",
        enabled=None, check_mode=False, force=False):
    """
    Duplicate and create an authentication policy
    """
    if force is True or _check(isamAppliance, name=name) is False:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            warnings = []
            json_data = {
                "name": name,
                "description": description,
                "policy": policy,
                "uri": uri,
                "dialect": dialect
            }
            if enabled is not None:
                if tools.version_compare(isamAppliance.facts["version"], "9.0.2.1") < 0:
                    warnings.append(
                        "Appliance is at version: {0}. Enabled parameter not supported unless atleast 9.0.2.1. Ignoring value.".format(
                            isamAppliance.facts["version"]))
                else:
                    json_data["enabled"] = enabled
            return isamAppliance.invoke_post(
                "Duplicate and create an authentication policy", module_uri, json_data,
                requires_modules=requires_modules, requires_version=requires_version, warnings=warnings)

    return isamAppliance.create_return_object()


def delete(isamAppliance, id=None, name=None, check_mode=False, force=False):
    """
    Delete an authentication policy
    """
    if id == None and name == None:
        from ibmsecurity.appliance.ibmappliance import IBMError
        raise IBMError("999", "Either id or name attribute must be provided")

    if force is True or _check(isamAppliance, id=id, name=name) is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            if id == None:
                ret_obj = get(isamAppliance, name)
                id = ret_obj['data']['id']
            return isamAppliance.invoke_delete(
                "Delete an authentication policy",
                "{0}/{1}".format(module_uri, id), requires_modules=requires_modules, requires_version=requires_version)

    return isamAppliance.create_return_object()


def update(isamAppliance, name, policy, uri, description="",
           dialect="urn:ibm:security:authentication:policy:1.0:schema", enabled=None, check_mode=False, force=False):
    """
    Update a specified authentication policy
    """
    warnings = []
    needs_update = False
    json_data = {
        "name": name,
        "description": description,
        "policy": policy,
        "uri": uri,
        "dialect": dialect
    }
    if enabled is not None:
        if tools.version_compare(isamAppliance.facts["version"], "9.0.2.1") < 0:
            warnings.append(
                "Appliance is at version: {0}. Enabled parameter not supported unless atleast 9.0.2.1. Ignoring value.".format(
                    isamAppliance.facts["version"]))
        else:
            json_data["enabled"] = enabled
    if force is not True:
        try:
            ret_obj = get(isamAppliance, name)
            id = ret_obj['data']['id']
            del ret_obj['data']['id']
            del ret_obj['data']['datecreated']
            del ret_obj['data']['lastmodified']
            del ret_obj['data']['userlastmodified']
            del ret_obj['data']['predefined']
            import ibmsecurity.utilities.tools
            exist_data = ibmsecurity.utilities.tools.json_sort(ret_obj['data'])
            new_data = ibmsecurity.utilities.tools.json_sort(json_data)
            logger.debug("Existing Data: {0}".format(exist_data))
            logger.debug("Provided Data: {0}".format(new_data))
            if exist_data != new_data:
                needs_update = True
        except:
            pass

    if force is True or needs_update is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_put(
                "Update a specified authentication policy",
                "{0}/{1}".format(module_uri, id), json_data, requires_modules=requires_modules,
                requires_version=requires_version, warnings=warnings)

    return isamAppliance.create_return_object()


def _check(isamAppliance, id=None, name=None):
    """
    Check if API Protection Definition already exists
    """
    ret_obj = get_all(isamAppliance)

    for obj in ret_obj['data']:
        if (id is not None and obj['id'] == id) or (name is not None and obj['name'] == name):
            return True

    return False


def search(isamAppliance, name, check_mode=False, force=False):
    """
    Retrieve the id for a given policy name
    """
    ret_obj = isamAppliance.create_return_object()
    ret_obj_all = get_all(isamAppliance)

    for obj in ret_obj_all['data']:
        if obj['name'] == name:
            ret_obj['data'] = obj['id']
            break

    return ret_obj


def activate(isamAppliance, name, enabled=True, check_mode=False, force=False):
    """
    Enable or disable a policy
    """
    warnings = []
    if tools.version_compare(isamAppliance.facts["version"], "9.0.2.1") < 0:
        warnings.append(
            "Appliance is at version: {0}. Enabled parameter not supported unless atleast 9.0.2.1. Ignoring value.".format(
                isamAppliance.facts["version"]))
    else:
        ret_obj = get(isamAppliance, name=name)
        if force or ret_obj['data']['enabled'] != enabled:
            if check_mode is True:
                return isamAppliance.create_return_object(changed=True, warnings=warnings)
            else:
                return update(isamAppliance, name=name, policy=ret_obj['data']['policy'], uri=ret_obj['data']['uri'],
                              description=ret_obj['data']['description'], dialect=ret_obj['data']['dialect'],
                              enabled=enabled)

    return isamAppliance.create_return_object(warnings=warnings)


def compare(isamAppliance1, isamAppliance2):
    """
    Compare Authentication Policies between two appliances
    """
    ret_obj1 = get_all(isamAppliance1)
    ret_obj2 = get_all(isamAppliance2)

    for obj in ret_obj1['data']:
        ret_obj = _get(isamAppliance1, obj['id'])
        obj['policy'] = ret_obj['data']['policy']
        del obj['id']
        del obj['datecreated']
        del obj['lastmodified']
        del obj['userlastmodified']
    for obj in ret_obj2['data']:
        ret_obj = _get(isamAppliance2, obj['id'])
        obj['policy'] = ret_obj['data']['policy']
        del obj['id']
        del obj['datecreated']
        del obj['lastmodified']
        del obj['userlastmodified']

    return tools.json_compare(ret_obj1, ret_obj2,
                              deleted_keys=['id', 'datecreated', 'lastmodified', 'userlastmodified'])



import logging
from ibmsecurity.utilities import tools

logger = logging.getLogger(__name__)

try:
    basestring
except NameError:
    basestring = (str, bytes)


def get(isamAppliance, state_id, check_mode=False, force=False):
    """
    Get a grant
    """
    return isamAppliance.invoke_get("Get a grant",
                                    "/iam/access/v8/grants/{0}".format(state_id))


def get_all(isamAppliance, check_mode=False, force=False):
    """
    Get all grants
    """
    return isamAppliance.invoke_get("Get all grants",
                                    "/iam/access/v8/grants")


def get_users(isamAppliance, sortBy=None, filter=None, check_mode=False, force=False):
    """
    Get all users that have grants
    """
    return isamAppliance.invoke_get("Get all users that have grants",
                                    "/iam/access/v8/grants/userIds{0}".format(
                                        tools.create_query_string(sortBy=sortBy, filter=filter)))


def set(isamAppliance, state_id, attributes=[], isEnabled=True, check_mode=False,
        force=False):
    """
    Update a specified grant

    NOTE: Unable to update attributes, getting an error like so:
        {"result":"FBTRBA0100E The action: UPDATE failed because the resource [https:\/\/<hostname>:443\/iam\/access\/v8\/grants\/uuid70c19d9-0158-1c78-8de9-cb87b95521e8] was not found."}
    """
    if (isinstance(attributes, basestring)):
        import ast
        attributes = ast.literal_eval(attributes)
    if (isinstance(isEnabled, basestring)):
        import ast
        isEnabled = ast.literal_eval(isEnabled)

    if force is True or _check(isamAppliance, state_id, isEnabled, attributes) is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_put("Update a specified grant",
                                            "/iam/access/v8/grants/{0}".format(state_id),
                                            {
                                                'isEnabled': isEnabled,
                                                'attributes': attributes
                                            })

    return isamAppliance.create_return_object()


def delete(isamAppliance, state_id, check_mode=False, force=False):
    """
    Delete a grant
    """
    if force is True or _check(isamAppliance, state_id) is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_delete("Delete a grant",
                                               "/iam/access/v8/grants/{0}".format(state_id))

    return isamAppliance.create_return_object()


def _check(isamAppliance, state_id, isEnabled=None, attributes=None):
    try:
        ret_obj = get(isamAppliance, state_id)
        if ret_obj['data']['id'] == state_id:
            if isEnabled is None and attributes is None:
                return True
            else:
                if ret_obj['data']['isEnabled'] != isEnabled:
                    return True
                import ibmsecurity.utilities.tools
                if ibmsecurity.utilities.tools.json_sort(
                        ret_obj['data']['attributes']) != ibmsecurity.utilities.tools.json_sort(attributes):
                    return True
    except:
        pass

    return False



import logging
import os.path
import shutil

from ibmsecurity.utilities.tools import get_random_temp_dir, files_same

logger = logging.getLogger(__name__)
uri = "/wga/reverseproxy"
requires_modules = ["wga"]
requires_version = None


def get(isamAppliance, instance_id, check_mode=False, force=False):
    """
    Retrieving a tracing configuration file

    """
    return isamAppliance.invoke_get("Retrieving a tracing configuration file",
                                    "{0}/{1}/tracing_configuration".format(uri, instance_id),
                                    requires_modules=requires_modules, requires_version=requires_version)


def export_file(isamAppliance, instance_id, filepath, check_mode=False, force=False):
    """
    Exporting a tracing configuration file
    """

    if os.path.exists(filepath) is True:
        logger.info("File '{0}' already exists.  Skipping export.".format(filepath))
        warnings = ["File '{0}' already exists.  Skipping export.".format(filepath)]
        return isamAppliance.create_return_object(warnings=warnings)

    if check_mode is True:
        return isamAppliance.create_return_object(changed=True)
    else:
        return isamAppliance.invoke_get_file(
            "Exporting a tracing configuration file",
            "{0}/{1}/tracing_configuration?export".format(uri, instance_id), filepath
        )

    return isamAppliance.create_return_object()


def update(isamAppliance, instance_id, contents, check_mode=False, force=False):
    """
    Updating tracing configuration file data - using contents string

    """
    update_required = False

    ret_obj = get(isamAppliance, instance_id)
    ret_contents = ret_obj['data']['contents']

    if ret_contents.strip() != contents.strip():
        update_required = True

    if force is True or update_required is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:

            return isamAppliance.invoke_put(
                "Updating tracing configuration file data - using contents string",
                "{0}/{1}/tracing_configuration".format(uri, instance_id),
                {
                    'file_contents': contents
                },
                requires_modules=requires_modules, requires_version=requires_version
            )

    return isamAppliance.create_return_object()


def import_file(isamAppliance, instance_id, filepath, check_mode=False, force=False):
    """
    Updating tracing configuration file data - using a file

    """

    if force is True or _check_import(isamAppliance=isamAppliance, id=instance_id, filepath=filepath) is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:

            return isamAppliance.invoke_put_files(
                "Updating tracing configuration file data - using a file",
                "{0}/{1}/tracing_configuration".format(uri, instance_id),
                [
                    {
                        'file_formfield': 'file',
                        'filename': filepath,
                        'mimetype': 'application/octet-stream'
                    }
                ],
                {},
                requires_modules=requires_modules, requires_version=requires_version
            )

    return isamAppliance.create_return_object()


def _check_import(isamAppliance, id, filepath):
    """
    Checks if the file to be imported is the same as the file that's already on the instance
    """
    tmpdir = get_random_temp_dir()
    tmp_original_file = os.path.join(tmpdir, os.path.basename("tempfile.txt"))

    export_file(isamAppliance, instance_id=id, filepath=tmp_original_file, check_mode=False, force=True)

    if files_same(tmp_original_file, filepath):
        logger.debug("files are the same, so we don't want to do anything")
        shutil.rmtree(tmpdir)
        return False
    else:
        logger.debug("files are different, so we return True to indicate the new file should be imported")
        shutil.rmtree(tmpdir)
        return True



import logging

logger = logging.getLogger(__name__)
requires_model="Appliance"


def get(isamAppliance, check_mode=False, force=False):
    """
    Retrieving the SSL certificate cluster replication status
    """
    return isamAppliance.invoke_get("Retrieving the SSL certificate cluster replication status",
                                    "/isam/ssl_certificates/?cluster=true", requires_model=requires_model)


def set(isamAppliance, replicating, check_mode=False, force=False):
    """
    Setting the SSL certificate cluster replication state
    """

    check_obj = _check(isamAppliance, replicating)
    if force is True or check_obj['value'] is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True, warnings=check_obj['warnings'])
        else:
            return isamAppliance.invoke_put("Setting the SSL certificate cluster replication state",
                                            "/isam/ssl_certificates/v1",
                                            {
                                                'replicating': replicating
                                            }, requires_model=requires_model)

    return isamAppliance.create_return_object(warnings=check_obj['warnings'])


def _check(isamAppliance, replicating):
    """
    Check the replicating status
    :param isamAppliance:
    :return:
    """

    check_obj = {'value': False, 'warnings':""}
    ret_obj = get(isamAppliance)
    check_obj['warnings']=ret_obj['warnings']

    if ret_obj['data'] != {}:
        if ret_obj['data']['replicating'] != replicating:
            check_obj['value'] = True
            return check_obj
        else:
            check_obj['value'] = False
            return check_obj

    return check_obj



import logging
from ibmsecurity.utilities import tools

logger = logging.getLogger(__name__)
requires_model = "Appliance"

def get_network(isamAppliance, application_interface, statistics_duration, check_mode=False, force=False):
    """
    Retrieving the Application Interface Statistics
    """
    return isamAppliance.invoke_get("Retrieving the Application Interface Statistics",
                                    "/analysis/interface_statistics.json{0}".format(
                                        tools.create_query_string(prefix=application_interface,
                                                                  timespan=statistics_duration)),requires_model=requires_model)


def get_rp_junction(isamAppliance, instance, date, duration, check_mode=False, force=False):
    """
    Retrieving junction average response times for a Reverse Proxy instance
    """
    return isamAppliance.invoke_get("Retrieving junction average response times for a Reverse Proxy instance",
                                    "/analysis/reverse_proxy_traffic/reqtime{0}".format(
                                        tools.create_query_string(date=date,
                                                                  duration=duration,
                                                                  instance=instance)),requires_model=requires_model)


def get_rp_health_summary(isamAppliance, check_mode=False, force=False):
    """
    Retrieving a summary of Reverse Proxy health
    """
    return isamAppliance.invoke_get("Retrieving a summary of Reverse Proxy health",
                                    "/wga/widgets/health.json",requires_model=requires_model)


def get_rp_throughput_summary(isamAppliance, date, duration, aspect, summary=None, check_mode=False, force=False):
    """
    Retrieving a summary of throughput for all Reverse Proxy instances
    """
    headers = {'Accept': 'application/json', 'range': 'items=0-24'}
    return isamAppliance.invoke_get_with_headers("Retrieving a summary of throughput for all Reverse Proxy instances",
                                                 "/analysis/reverse_proxy_traffic/throughput/{0}".format(
                                                     tools.create_query_string(summary=summary,
                                                                               date=date,
                                                                               duration=duration,
                                                                               aspect=aspect)), 
                                                                               requires_model=requires_model,
                                                                               headers=headers)


def get_rp_throughput(isamAppliance, instance, date, duration, check_mode=False, force=False):
    """
    Retrieving throughput records for a specific Reverse Proxy instance
    """
    return isamAppliance.invoke_get("Retrieving throughput records for a specific Reverse Proxy instance",
                                    "/analysis/reverse_proxy_traffic/throughput/{0}{1}".format(instance,
                                                                                               tools.create_query_string(
                                                                                                   date=date,
                                                                                                   duration=duration)),requires_model=requires_model)


def get_rp_traffic_summary(isamAppliance, instance, date, duration, aspect, summary=True, check_mode=False,
                           force=False):
    """
    Retrieving a summary of traffic by Junction or User-Agent on a Reverse Proxy instance
    """
    return isamAppliance.invoke_get(
        "Retrieving a summary of traffic by Junction or User-Agent on a Reverse Proxy instance",
        "/analysis/reverse_proxy_traffic/traffic/instance/{0}/{1}".format(instance,
                                                                          tools.create_query_string(summary=summary,
                                                                                                    date=date,
                                                                                                    duration=duration,
                                                                                                    aspect=aspect)),requires_model=requires_model)


def get_rp_traffic(isamAppliance, instance, date, duration, aspect, aspect_identifier, check_mode=False, force=False):
    """
    Retrieving a summary of traffic records for a specific Junction or User-Agent on a Reverse Proxy instance
    """
    return isamAppliance.invoke_get(
        "Retrieving a summary of traffic records for a specific Junction or User-Agent on a Reverse Proxy instance",
        "/analysis/reverse_proxy_traffic/traffic/instance/{0}/{1}/{2}{3}".format(instance, aspect, aspect_identifier,
                                                                                 tools.create_query_string(date=date,
                                                                                                           duration=duration)),requires_model=requires_model)


def get_rp_traffic_detail(isamAppliance, instance, date, duration, aspect, aspect_identifier, check_mode=False,
                          force=False):
    """
    Retrieving detailed traffic records for a specific Junction or User-Agent on a Reverse Proxy instance
    """
    return isamAppliance.invoke_get(
        "Retrieving detailed traffic records for a specific Junction or User-Agent on a Reverse Proxy instance",
        "/analysis/reverse_proxy_traffic/traffic/instance/{0}/{1}/{2}/{3}".format(instance, aspect, aspect_identifier,
                                                                                  tools.create_query_string(date=date,
                                                                                                            duration=duration,
                                                                                                            aspect=aspect)),requires_model=requires_model)


def get_rp_traffic_detail_aspect(isamAppliance, instance, date, duration, aspect, aspect_identifier, check_mode=False,
                                 force=False):
    """
    Retrieving detailed traffic records for a specific User-Agent on a specific junction in a Reverse Proxy instance
    """
    return isamAppliance.invoke_get(
        "Retrieving detailed traffic records for a specific User-Agent on a specific junction in a Reverse Proxy instance",
        "/analysis/reverse_proxy_traffic/traffic/instance/{0}/{1}/{2}/{1}/{2}{3}".format(instance, aspect,
                                                                                         aspect_identifier,
                                                                                         tools.create_query_string(
                                                                                             date=date,
                                                                                             duration=duration)),requires_model=requires_model)


def get_rp_waf_events(isamAppliance, instance, date, duration, type, check_mode=False,
                      force=False):
    """
    Retrieving security action events for a Reverse Proxy instance
    """
    return isamAppliance.invoke_get(
        "Retrieving security action events for a Reverse Proxy instance",
        "/analysis/reverse_proxy_traffic/pam_events{0}".format(
            tools.create_query_string(
                date=date,
                duration=duration,
                instance=instance,
                type=type)),requires_model=requires_model)


def get_cpu(isamAppliance, statistics_duration, check_mode=False, force=False):
    """
    Retrieving the CPU Usage Statistics
    """
    return isamAppliance.invoke_get(
        "Retrieving the CPU Usage Statistics",
        "/statistics/systems/cpu.json{0}".format(
            tools.create_query_string(
                timespan=statistics_duration)),requires_model=requires_model)


def get_memory(isamAppliance, statistics_duration, check_mode=False, force=False):
    """
    Retrieving the Storage Usage Statistics
    """
    return isamAppliance.invoke_get(
        "Retrieving the Memory Usage Statistics",
        "/statistics/systems/memory.json{0}".format(
            tools.create_query_string(
                timespan=statistics_duration)),requires_model=requires_model)


def get_storage(isamAppliance, statistics_duration, check_mode=False, force=False):
    """
    Retrieving the Storage Usage Statistics
    """
    return isamAppliance.invoke_get(
        "Retrieving the Storage Usage Statistics",
        "/statistics/systems/storage.json{0}".format(
            tools.create_query_string(
                timespan=statistics_duration)),requires_model=requires_model)



import logging
import ibmsecurity.utilities.tools

logger = logging.getLogger(__name__)


def get(isdsAppliance, check_mode=False, force=False):
    """
    Retrieving the administrator settings
    """
    return isdsAppliance.invoke_get("Retrieving the administrator settings", "/admin_cfg")


def set_pw(isdsAppliance, newPassword, oldPassword, sessionTimeout="30", check_mode=False, force=False):
    """
    Set password for admin user (super user for appliance)
    """
    warnings = ["Password change requested - cannot query existing password for idempotency check."]
    if check_mode is True:
        return isdsAppliance.create_return_object(changed=True, warnings=warnings)
    else:
        json_data = {
            "oldPassword": oldPassword,
            "newPassword": newPassword,
            "confirmPassword": newPassword,
            "sessionTimeout": sessionTimeout
        }

        return isdsAppliance.invoke_put("Setting admin password", "/admin_cfg", json_data, warnings=warnings)


def set(isdsAppliance, oldPassword=None, newPassword=None, sessionTimeout=30, check_mode=False, force=False):
    """
    Updating the administrator settings
    """
    warnings = []
    if force is False:
        update_required, warnings, json_data = _check(isdsAppliance, oldPassword, newPassword, sessionTimeout, warnings)

    if force is True or update_required is True:
        if check_mode is True:
            return isdsAppliance.create_return_object(changed=True, warnings=warnings)
        else:
            return isdsAppliance.invoke_put(
                "Updating the administrator settings",
                "/admin_cfg", json_data, warnings=warnings)

    return isdsAppliance.create_return_object(changed=False, warnings=warnings)


def _check(isdsAppliance, oldPassword, newPassword, sessionTimeout, warnings):
    """
    Check whether target key has already been set with the value
    :param isdsAppliance:
    :param key:
    :param value:
    :return: True/False
    """
    ret_obj = get(isdsAppliance)

    json_data = {
        "sessionTimeout": sessionTimeout
    }
    if oldPassword is not None:
        json_data["oldPassword"] = oldPassword
        if newPassword is None:
            warnings.append("Please provide new password, when old password is specified.")
    if newPassword is not None:
        warnings.append("Password change requested - cannot query existing password for idempotency check.")
        json_data["newPassword"] = newPassword
        json_data["confirmPassword"] = newPassword
        if oldPassword is None:
            warnings.append("Please provide old password, when new password is specified.")

    if ibmsecurity.utilities.tools.json_sort(json_data) != ibmsecurity.utilities.tools.json_sort(ret_obj['data']):
        logger.debug("Admin Settings are found to be different. See following JSON for difference.")
        logger.debug("New JSON: {0}".format(ibmsecurity.utilities.tools.json_sort(json_data)))
        logger.debug("Old JSON: {0}".format(ibmsecurity.utilities.tools.json_sort(ret_obj['data'])))
        return True, warnings, json_data
    else:  # No changes required
        return False, warnings, json_data


def compare(isdsAppliance1, isdsAppliance2):
    """
    Compare advanced tuning parameters between two appliances
    """
    ret_obj1 = get(isdsAppliance1)
    ret_obj2 = get(isdsAppliance2)

    return ibmsecurity.utilities.tools.json_compare(ret_obj1, ret_obj2)



import logging
import ibmsecurity.utilities.tools
from ibmsecurity.isam.aac.api_protection import definitions

logger = logging.getLogger(__name__)

# URI for this module
uri = "/wga/rsa_config"
requires_modules = ["wga"]
requires_version = None
requires_model = "Appliance"


def get(isamAppliance, check_mode=False, force=False):
    """
    Retrieve RSA Securid Configuration
    """
    return isamAppliance.invoke_get("Retrieve RSA Securid Configuration", uri, requires_modules=requires_modules,
                                    requires_version=requires_version, requires_model=requires_model)


def upload(isamAppliance, filename, check_mode=False, force=False):
    """
    Upload a RSA Securid Config file
    """

    srv_cfg_available, warnings = _check(isamAppliance)
    if warnings and warnings[0] != '':
        return isamAppliance.create_return_object(warnings=warnings)

    if force is True or srv_cfg_available is False:
        warnings = [
                "Idempotency check is only to see if there was a config already uploaded. Force upload to replace existing configuration."]

        if check_mode is True:
            return isamAppliance.create_return_object(changed=True, warnings=warnings)
        else:
            return isamAppliance.invoke_post_files(
                    "Upload a RSA Securid Config file",
                    "{0}/server_config".format(uri),
                    [
                        dict(file_formfield='server_config_file', filename=filename,
                             mimetype='application/octet-stream')
                    ],
                    {}, requires_modules=requires_modules,
                    requires_version=requires_version, warnings=warnings, requires_model=requires_model)

    return isamAppliance.create_return_object(warnings=warnings)


def _check(isamAppliance):
    ret_obj = get(isamAppliance)
    srv_cfg_available, warnings = False, ret_obj['warnings']

    if warnings == [] and ret_obj['data']['server_config'] == 'available':
        srv_cfg_available = True
    return srv_cfg_available, warnings


def test(isamAppliance, username, passcode, check_mode=False, force=False):
    """
    Test RSA Configuration with username/passcode
    """
    srv_cfg_available, warnings = _check(isamAppliance)

    if warnings and warnings[0] != '':
        return isamAppliance.create_return_object(warnings=warnings)

    if srv_cfg_available is False:
        return isamAppliance.create_return_object(warnings=["Valid configuration not found, test skipped."])
    else:
        ret_obj = isamAppliance.invoke_post("Test RSA Configuration with username/passcode", "{0}/test".format(uri),
                                            {
                                                'username': username,
                                                'passcode': passcode
                                            },
                                            requires_modules=requires_modules,
                                            requires_version=requires_version,
                                            ignore_error=True,
                                            requires_model=requires_model)
        if ret_obj['changed'] is True:
            ret_obj['changed'] = False

        return ret_obj


def delete(isamAppliance, check_mode=False, force=False):
    """
    Deleting or Clear RSA Securid Configuration
    """
    srv_cfg_available, warnings = _check(isamAppliance)

    if force is True or srv_cfg_available is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True, warnings=warnings)
        else:
            return isamAppliance.invoke_delete("Deleting or Clear RSA Securid Configuration",
                                               "{0}/server_config".format(uri),
                                               requires_modules=requires_modules,
                                               requires_version=requires_version,
                                               requires_model=requires_model)

    return isamAppliance.create_return_object(warnings=warnings)


def clear(isamAppliance, check_mode=False, force=False):
    """
    Clear the node secret file

    :param isamAppliance:
    :param check_mode:
    :param force:
    :return:
    """
    # TODO: This function has not been tested.  Please open an issue on GitHub if you find a problem.
    srv_cfg_available, warnings = _check(isamAppliance)

    if force is True or srv_cfg_available is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True, warnings=warnings)
        else:
            return isamAppliance.invoke_delete("Clear the node secret file",
                                               "{0}/node_secret".format(uri),
                                               requires_modules=requires_modules,
                                               requires_version=requires_version,
                                               requires_model=requires_model)

    return isamAppliance.create_return_object(warnings=warnings)



import logging
import ibmsecurity.utilities.tools

logger = logging.getLogger(__name__)


def get(isdsAppliance, ip_addr, check_mode=False, force=False):
    """
    Get hostnames for a host record
    """
    return isdsAppliance.invoke_get("Retrieving hostnames for a host record",
                                    "/host_records/" + ip_addr + "/hostnames")


def add(isdsAppliance, hostname, ip_addr, check_mode=False, force=False):
    """
    Add hostname to host record
    """
    if force is True or _check(isdsAppliance, hostname, ip_addr) is False:
        if check_mode is True:
            return isdsAppliance.create_return_object(changed=True)
        else:
            return isdsAppliance.invoke_post(
                "Adding hostname to host record",
                "/host_records/" + ip_addr + "/hostnames",
                {
                    'name': hostname
                })

    return isdsAppliance.create_return_object()


def delete(isdsAppliance, hostname, ip_addr, check_mode=False, force=False):
    """
    Delete a hostname from a host record
    """
    if force is True or _check(isdsAppliance, hostname, ip_addr) is True:
        if check_mode is True:
            return isdsAppliance.create_return_object(changed=True)
        else:
            return isdsAppliance.invoke_delete(
                "Removing hostname",
                "/host_records/" + ip_addr + "/hostnames/" + hostname)

    return isdsAppliance.create_return_object()


def _check(isdsAppliance, hostname, ip_addr, check_mode=False, force=False):
    """
    Check whether hostname has been defined
    """
    ret_obj = get(isdsAppliance, ip_addr)

    for names in ret_obj['data']:
        if names['name'] == hostname:
            logger.info("Hostname record exists")
            return True

    return False



import logging
import ibmsecurity.utilities.tools

logger = logging.getLogger(__name__)
requires_model="Appliance"


def get(isamAppliance, check_mode=False, force=False):
    """
    Retrieving the current date and time settings
    """
    return isamAppliance.invoke_get("Retrieving the current date and time settings",
                                    "/time_cfg", requires_model=requires_model)


def get_timezones(isamAppliance, check_mode=False, force=False):
    """
    Retrieving the list of valid timezones
    """
    return isamAppliance.invoke_get("Retrieving the list of valid timezones",
                                    "/time_cfg/I18nTimezone", requires_model=requires_model)


def set(isamAppliance, ntpServers="", timeZone="America/New_York", enableNtp=False, dateTime=None, check_mode=False, force=False):
    """
    Update date/time settings (set NTP server and timezone)
    """
    
    if dateTime is None:
        dateTime = "0000-00-00 00:00:00"

    check_value, warnings = _check(isamAppliance=isamAppliance, timeZone=timeZone, ntpServers=ntpServers, enableNtp=enableNtp)
    if force is True or check_value is False:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True, warnings=warnings)
        else:
            return isamAppliance.invoke_put(
                "Setting date/time settings (NTP)",
                "/time_cfg",
                {
                    "dateTime": dateTime,
                    "timeZone": timeZone,
                    "enableNtp": enableNtp,
                    "ntpServers": ntpServers
                },
            requires_model=requires_model
            )

    return isamAppliance.create_return_object(warnings=warnings)


def disable(isamAppliance, check_mode=False, force=False):
    """
    Disable NTP settings, leaves all other settings intact
    """

    ret_obj = get(isamAppliance)
    warnings = ret_obj['warnings']

    if warnings != []:
        if "Docker" in warnings[0]:
            return isamAppliance.create_return_object(warnings=warnings)

    if force is True or ret_obj['data']['ntpConfig']['enableNtp'] == True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True, warnings=warnings)
        else:
            return isamAppliance.invoke_put(
                "Setting date/time settings (NTP)",
                "/time_cfg",
                {
                    'dateTime': ret_obj['data']['dateTime'],
                    'timeZone': ret_obj['data']['timeZone'],
                    'enableNtp': False,
                    'ntpServers': ','.join(ns['ntpServer'] for ns in ret_obj['data']['ntpConfig']['ntpServers'])
                },
                requires_model=requires_model
            )

    return isamAppliance.create_return_object(warnings=warnings)


def _check(isamAppliance, timeZone, ntpServers, enableNtp):
    """
    Check if NTP is already set for syncing date/time
    If timezone or ntpservers provided then also check if those values match what is currently set
    """
    ret_obj = get(isamAppliance)
    warnings=ret_obj['warnings']

    if warnings != []:
        if "Docker" in warnings[0]:
            return True, warnings

    if ret_obj['data']['ntpConfig']['enableNtp'] != enableNtp:
        return False, warnings

    if timeZone != ret_obj['data']['timeZone']:
        logger.info("Existing timeZone is different")
        return False, warnings

    if ntpServers != None:
        existing_ntpServers = list()
        for ntps in ret_obj['data']['ntpConfig']['ntpServers']:
            existing_ntpServers.append(ntps['ntpServer'])
        logger.debug(str(sorted(existing_ntpServers)))
        logger.debug(str(sorted(ntpServers.split(','))))
        if sorted(ntpServers.split(',')) != sorted(existing_ntpServers):
            logger.debug("Existing ntpServers are different")
            return False, warnings

    return True, warnings


def compare(isamAppliance1, isamAppliance2):
    """
    Compare date/time settings between two appliances
    """
    ret_obj1 = get(isamAppliance1)
    ret_obj2 = get(isamAppliance2)

    if ret_obj1['warnings'] != []:
        if 'Docker' in ret_obj1['warnings'][0]:
            return isamAppliance1.create_return_object(warnings=ret_obj1['warnings'])

    if ret_obj2['warnings'] != []:
        if 'Docker' in ret_obj2['warnings'][0]:
            return isamAppliance2.create_return_object(warnings=ret_obj2['warnings'])

    # Ignore actual date / time on servers - they should be same if synced correctly
    if 'dateTime' in ret_obj1['data']:
        del ret_obj1['data']['dateTime']

    if 'dateTime' in ret_obj2['data']:
        del ret_obj2['data']['dateTime']

    return ibmsecurity.utilities.tools.json_compare(ret_obj1, ret_obj2, deleted_keys=['datetime'])
